#
# SPDX-FileCopyrightText: Copyright (c) provide.io llc. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#


from __future__ import annotations

from decimal import Decimal
from typing import Any, cast

from pyvider.cty import CtyBool, CtyNumber, CtyString, CtyValue
from pyvider.cty.config.defaults import (
    COMPARISON_OPS_MAP,
    ERR_ALL_ARGS_SAME_TYPE,
    ERR_CANNOT_COMPARE,
    ERR_MIN_ONE_ARG,
)
from pyvider.cty.exceptions import CtyFunctionError
from pyvider.cty.values.markers import RefinedUnknownValue


def equal(a: CtyValue[Any], b: CtyValue[Any]) -> CtyValue[Any]:
    if a.is_unknown or b.is_unknown:
        return CtyValue.unknown(CtyBool())
    return CtyBool().validate(a == b)


def not_equal(a: CtyValue[Any], b: CtyValue[Any]) -> CtyValue[Any]:
    if a.is_unknown or b.is_unknown:
        return CtyValue.unknown(CtyBool())
    return CtyBool().validate(a != b)


def _compare(a: CtyValue[Any], b: CtyValue[Any], op: str) -> CtyValue[Any]:  # noqa: C901
    if a.is_null or b.is_null:
        return CtyValue.unknown(CtyBool())

    # Handle refined unknown comparisons
    if a.is_unknown or b.is_unknown:
        ref_a = a.value if isinstance(a.value, RefinedUnknownValue) else None
        ref_b = b.value if isinstance(b.value, RefinedUnknownValue) else None

        # Case 1: One is known, one is refined unknown
        if a.is_unknown and not b.is_unknown and ref_a:
            b_val = cast(Decimal, b.value)
            if ref_a.number_upper_bound:
                upper, inclusive = ref_a.number_upper_bound
                if b_val > upper or (b_val == upper and not inclusive):
                    if op in (">", ">="):
                        return CtyBool().validate(False)
                    if op in ("<", "<="):
                        return CtyBool().validate(True)
            if ref_a.number_lower_bound:
                lower, inclusive = ref_a.number_lower_bound
                if b_val < lower or (b_val == lower and not inclusive):
                    if op in ("<", "<="):
                        return CtyBool().validate(False)
                    if op in (">", ">="):
                        return CtyBool().validate(True)
        elif b.is_unknown and not a.is_unknown and ref_b:
            a_val = cast(Decimal, a.value)
            if ref_b.number_upper_bound:
                upper, inclusive = ref_b.number_upper_bound
                if a_val > upper or (a_val == upper and not inclusive):
                    if op in ("<", "<="):
                        return CtyBool().validate(False)
                    if op in (">", ">="):
                        return CtyBool().validate(True)
            if ref_b.number_lower_bound:
                lower, inclusive = ref_b.number_lower_bound
                if a_val < lower or (a_val == lower and not inclusive):
                    if op in (">", ">="):
                        return CtyBool().validate(False)
                    if op in ("<", "<="):
                        return CtyBool().validate(True)
        # Case 2: Both are refined unknowns
        elif a.is_unknown and b.is_unknown and ref_a and ref_b:
            if ref_a.number_upper_bound and ref_b.number_lower_bound:
                a_upper, a_inc = ref_a.number_upper_bound
                b_lower, b_inc = ref_b.number_lower_bound
                if a_upper < b_lower or (a_upper == b_lower and not (a_inc and b_inc)):
                    if op in ("<", "<="):
                        return CtyBool().validate(True)
                    if op in (">", ">="):
                        return CtyBool().validate(False)
            if ref_a.number_lower_bound and ref_b.number_upper_bound:
                a_lower, a_inc = ref_a.number_lower_bound
                b_upper, b_inc = ref_b.number_upper_bound
                if a_lower > b_upper or (a_lower == b_upper and not (a_inc and b_inc)):
                    if op in (">", ">="):
                        return CtyBool().validate(True)
                    if op in ("<", "<="):
                        return CtyBool().validate(False)

        return CtyValue.unknown(CtyBool())

    # Handle known value comparisons
    if not isinstance(a.type, CtyNumber | CtyString) or not a.type.equal(b.type):
        error_message = ERR_CANNOT_COMPARE.format(type1=a.type.ctype, type2=b.type.ctype)
        raise CtyFunctionError(error_message)

    ops = COMPARISON_OPS_MAP
    return CtyBool().validate(ops[op](a.value, b.value))


def greater_than(a: CtyValue[Any], b: CtyValue[Any]) -> CtyValue[Any]:
    return _compare(a, b, ">")


def greater_than_or_equal_to(a: CtyValue[Any], b: CtyValue[Any]) -> CtyValue[Any]:
    return _compare(a, b, ">=")


def less_than(a: CtyValue[Any], b: CtyValue[Any]) -> CtyValue[Any]:
    return _compare(a, b, "<")


def less_than_or_equal_to(a: CtyValue[Any], b: CtyValue[Any]) -> CtyValue[Any]:
    return _compare(a, b, "<=")


def _partition_args(
    *args: CtyValue[Any],
) -> tuple[list[CtyValue[Any]], list[CtyValue[Any]]]:
    """Separate arguments into known and unknown values."""
    known_args, unknown_args = [], []
    for v in args:
        if v.is_unknown:
            unknown_args.append(v)
        elif not v.is_null:
            known_args.append(v)
    return known_args, unknown_args


def _validate_homogeneous_types(known_args: list[CtyValue[Any]], op: str) -> None:
    """Ensure all known arguments have compatible types."""
    if not known_args:
        return
    is_all_numbers = all(isinstance(v.type, CtyNumber) for v in known_args)
    is_all_strings = all(isinstance(v.type, CtyString) for v in known_args)
    if not (is_all_numbers or is_all_strings):
        error_message = ERR_ALL_ARGS_SAME_TYPE.format(op=op)
        raise CtyFunctionError(error_message)


def _find_extreme_value(known_args: list[CtyValue[Any]], op: str) -> CtyValue[Any] | None:
    """Find the extreme (min/max) value among known arguments."""
    if not known_args:
        return None
    if op == "max":
        result: CtyValue[Any] = max(known_args, key=lambda v: v.value)  # type: ignore[arg-type,return-value]
        return result
    else:
        result = min(known_args, key=lambda v: v.value)  # type: ignore[arg-type,return-value]
        return result


def _filter_dominated_unknowns(
    unknown_args: list[CtyValue[Any]], extreme_known: CtyValue[Any], op: str
) -> list[CtyValue[Any]]:
    """Remove unknown values that are definitely dominated by the extreme known value."""
    remaining_unknowns = []
    extreme_val = cast(Decimal, extreme_known.value)
    for unk in unknown_args:
        if isinstance(unk.value, RefinedUnknownValue):
            ref = unk.value
            if op == "max":
                if ref.number_upper_bound and (extreme_val >= ref.number_upper_bound[0]):
                    continue
            elif op == "min" and ref.number_lower_bound and (extreme_val <= ref.number_lower_bound[0]):
                continue
        remaining_unknowns.append(unk)
    return remaining_unknowns


def _multi_compare(*args: CtyValue[Any], op: str) -> CtyValue[Any]:
    if not args:
        error_message = ERR_MIN_ONE_ARG.format(op=op)
        raise CtyFunctionError(error_message)

    known_args, unknown_args = _partition_args(*args)

    if not known_args and not unknown_args:
        return CtyValue.null(args[0].type)

    _validate_homogeneous_types(known_args, op)
    extreme_known = _find_extreme_value(known_args, op)

    if extreme_known:
        unknown_args = _filter_dominated_unknowns(unknown_args, extreme_known, op)

    if not unknown_args:
        return extreme_known if extreme_known else CtyValue.null(args[0].type)
    if not known_args and len(unknown_args) == 1:
        return unknown_args[0]
    return CtyValue.unknown(args[0].type)


def max_fn(*args: CtyValue[Any]) -> CtyValue[Any]:
    return _multi_compare(*args, op="max")


def min_fn(*args: CtyValue[Any]) -> CtyValue[Any]:
    return _multi_compare(*args, op="min")


# 🌊🪢🔚



#
# SPDX-FileCopyrightText: Copyright (c) provide.io llc. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#


from __future__ import annotations

from collections.abc import Iterator
from typing import (
    TYPE_CHECKING,
    Any,
    Generic,
    Self,
    TypeVar,
)

from attrs import define, evolve, field

from pyvider.cty.config.defaults import (
    ERR_CANNOT_COMPARE_CTYVALUE_WITH,
    ERR_CANNOT_COMPARE_DIFFERENT_TYPES,
    ERR_CANNOT_COMPARE_NULL_UNKNOWN,
    ERR_CANNOT_GET_LENGTH_UNKNOWN_VALUE,
    ERR_CANNOT_GET_RAW_VALUE_UNKNOWN,
    ERR_CANNOT_INDEX_UNKNOWN_NULL_VALUE,
    ERR_CANNOT_ITERATE_UNKNOWN_VALUE,
    ERR_UNHASHABLE_TYPE,
    ERR_VALUE_TYPE_NO_LEN,
    ERR_VALUE_TYPE_NOT_COMPARABLE,
    ERR_VALUE_TYPE_NOT_ITERABLE,
    ERR_VALUE_TYPE_NOT_SUBSCRIPTABLE,
)
from pyvider.cty.values.markers import UNREFINED_UNKNOWN

T = TypeVar("T", covariant=True)

if TYPE_CHECKING:
    from pyvider.cty.types import CtyType


@define(frozen=True, slots=True)
class CtyValue(Generic[T]):
    vtype: CtyType[T] = field()
    value: object | None = field(default=None)
    is_unknown: bool = field(default=False)
    is_null: bool = field(default=False)
    marks: frozenset[Any] = field(factory=frozenset)

    def __attrs_post_init__(self) -> None:
        from pyvider.cty.types import CtyDynamic

        if isinstance(self.vtype, CtyDynamic) and isinstance(self.value, CtyValue):
            object.__setattr__(self, "is_unknown", self.value.is_unknown)
            object.__setattr__(self, "is_null", self.value.is_null)

        if self.is_unknown and self.is_null:
            object.__setattr__(self, "is_null", False)
        elif self.is_null and self.value is not None:
            object.__setattr__(self, "value", None)

    @property
    def type(self) -> CtyType[T]:
        return self.vtype

    @property
    def raw_value(self) -> object | None:
        if self.is_unknown:
            error_message = ERR_CANNOT_GET_RAW_VALUE_UNKNOWN
            raise ValueError(error_message)
        if self.is_null:
            return None
        from ..conversion.adapter import cty_to_native

        return cty_to_native(self)  # type: ignore

    def _canonical_sort_key(self) -> tuple[Any, ...]:
        from ..types import (
            CtyBool,
            CtyCapsule,
            CtyList,
            CtyMap,
            CtyNumber,
            CtyObject,
            CtySet,
            CtyString,
            CtyTuple,
        )

        if self.is_null:
            return (0,)
        if self.is_unknown:
            return (1,)

        type_rank = self.type._type_order
        key_prefix = (2, type_rank)

        if isinstance(self.type, CtyBool | CtyNumber | CtyString):
            return (*key_prefix, self.value)

        if (
            isinstance(self.type, CtyList | CtyTuple)
            and self.value is not None
            and hasattr(self.value, "__iter__")
        ):
            return (*key_prefix, *(v._canonical_sort_key() for v in self.value))

        if isinstance(self.type, CtySet) and self.value is not None and hasattr(self.value, "__iter__"):
            sorted_elements = sorted(self.value, key=lambda v: v._canonical_sort_key())
            return (*key_prefix, *(v._canonical_sort_key() for v in sorted_elements))

        if (
            isinstance(self.type, CtyMap | CtyObject)
            and self.value is not None
            and hasattr(self.value, "items")
        ):
            sorted_items = sorted(self.value.items())
            return (
                *key_prefix,
                *((k, v._canonical_sort_key()) for k, v in sorted_items),
            )

        if isinstance(self.type, CtyCapsule):
            return (*key_prefix, repr(self.value))

        # Fallback for any other type
        return (*key_prefix, repr(self.value))

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, CtyValue):
            return NotImplemented
        from ..types import CtyCapsuleWithOps

        if isinstance(self.type, CtyCapsuleWithOps) and self.type.equal(other.type) and self.type.equal_fn:
            return self.type.equal_fn(self.value, other.value)

        return (
            self.type.equal(other.type)
            and self.is_unknown == other.is_unknown
            and self.is_null == other.is_null
            and self.marks == other.marks
            and self.value == other.value
        )

    def _check_comparable(self, other: object) -> CtyValue[Any]:
        from ..types import CtyNumber, CtyString

        if not isinstance(other, CtyValue):
            error_message = ERR_CANNOT_COMPARE_CTYVALUE_WITH.format(type_name=type(other).__name__)
            raise TypeError(error_message)
        if self.is_unknown or self.is_null or other.is_unknown or other.is_null:
            error_message = ERR_CANNOT_COMPARE_NULL_UNKNOWN
            raise TypeError(error_message)
        if not self.type.equal(other.type):
            error_message = ERR_CANNOT_COMPARE_DIFFERENT_TYPES.format(type1=self.type, type2=other.type)
            raise TypeError(error_message)
        if not isinstance(self.type, CtyNumber | CtyString):
            error_message = ERR_VALUE_TYPE_NOT_COMPARABLE.format(type=self.type)
            raise TypeError(error_message)
        return other

    def __lt__(self, other: object) -> bool:
        other_val = self._check_comparable(other)
        if hasattr(self.value, "__lt__"):
            return bool(self.value < other_val.value)
        error_message = ERR_VALUE_TYPE_NOT_COMPARABLE.format(type=self.type)
        raise TypeError(error_message)

    def __le__(self, other: object) -> bool:
        other_val = self._check_comparable(other)
        if hasattr(self.value, "__le__"):
            return bool(self.value <= other_val.value)
        error_message = ERR_VALUE_TYPE_NOT_COMPARABLE.format(type=self.type)
        raise TypeError(error_message)

    def __gt__(self, other: object) -> bool:
        other_val = self._check_comparable(other)
        if hasattr(self.value, "__gt__"):
            return bool(self.value > other_val.value)
        error_message = ERR_VALUE_TYPE_NOT_COMPARABLE.format(type=self.type)
        raise TypeError(error_message)

    def __ge__(self, other: object) -> bool:
        other_val = self._check_comparable(other)
        if hasattr(self.value, "__ge__"):
            return bool(self.value >= other_val.value)
        error_message = ERR_VALUE_TYPE_NOT_COMPARABLE.format(type=self.type)
        raise TypeError(error_message)

    def __contains__(self, item: Any) -> bool:
        if self.is_unknown or self.is_null:
            return False
        if hasattr(self.value, "__contains__"):
            return item in self.value
        return bool(self.value == item)

    def __bool__(self) -> bool:
        from pyvider.cty.types import CtyDynamic

        if self.is_unknown or self.is_null:
            return False
        if isinstance(self.vtype, CtyDynamic) and isinstance(self.value, CtyValue):
            return bool(self.value)
        return True

    def __len__(self) -> int:
        from pyvider.cty.types import CtyDynamic, CtyList, CtyMap, CtySet, CtyTuple

        if self.is_unknown:
            error_message = ERR_CANNOT_GET_LENGTH_UNKNOWN_VALUE
            raise TypeError(error_message)
        if isinstance(self.vtype, CtyDynamic) and isinstance(self.value, CtyValue):
            return len(self.value)
        if self.is_null:
            return 0
        if isinstance(self.vtype, CtyList | CtyMap | CtySet | CtyTuple) and hasattr(self.value, "__len__"):
            return len(self.value)
        error_message = ERR_VALUE_TYPE_NO_LEN.format(type_name=self.vtype.__class__.__name__)
        raise TypeError(error_message)

    def __iter__(self) -> Iterator[Any]:
        from pyvider.cty.types import CtyList, CtyMap, CtySet, CtyTuple

        if self.is_unknown:
            error_message = ERR_CANNOT_ITERATE_UNKNOWN_VALUE
            raise TypeError(error_message)
        if self.is_null:
            return iter([])
        if isinstance(self.vtype, CtyList | CtySet | CtyTuple) and hasattr(self.value, "__iter__"):
            return iter(self.value)
        if isinstance(self.vtype, CtyMap) and hasattr(self.value, "values"):
            return iter(self.value.values())

        error_message = ERR_VALUE_TYPE_NOT_ITERABLE.format(type_name=self.vtype.__class__.__name__)
        raise TypeError(error_message)

    def __getitem__(self, key: Any) -> CtyValue[Any]:
        from ..types import CtyList, CtyMap, CtyObject, CtyTuple

        if self.is_unknown or self.is_null:
            error_message = ERR_CANNOT_INDEX_UNKNOWN_NULL_VALUE
            raise TypeError(error_message)
        if isinstance(self.vtype, CtyObject):
            if not isinstance(key, str):
                raise TypeError(f"Object attribute name must be a string, got {type(key).__name__}")
            return self.vtype.get_attribute(self, key)
        if isinstance(self.vtype, CtyList):
            if not isinstance(self.value, list | tuple):
                raise TypeError(f"CtyList value is not a list/tuple, but {type(self.value).__name__}")
            if isinstance(key, slice):
                return CtyValue(vtype=self.vtype, value=tuple(self.value[key]))
            return self.vtype.element_at(self, key)
        if isinstance(self.vtype, CtyTuple):
            return self.vtype.element_at(self, key)
        if isinstance(self.vtype, CtyMap):
            return self.vtype.get(self, key)  # type: ignore[arg-type]
        error_message = ERR_VALUE_TYPE_NOT_SUBSCRIPTABLE.format(type_name=self.vtype.__class__.__name__)
        raise TypeError(error_message)

    def __hash__(self) -> int:
        from pyvider.cty.types import (
            CtyCapsuleWithOps,
            CtyList,
            CtyMap,
            CtyObject,
            CtySet,
        )

        if isinstance(self.type, CtyCapsuleWithOps) and self.type.hash_fn:
            return self.type.hash_fn(self.value)

        if isinstance(self.vtype, CtyList | CtySet | CtyMap | CtyObject):
            error_message = ERR_UNHASHABLE_TYPE.format(vtype=self.vtype.ctype)
            raise TypeError(error_message)

        if self.is_unknown or self.is_null:
            return hash((self.vtype, self.is_unknown, self.is_null, self.marks))
        return hash((self.vtype, self.is_unknown, self.is_null, self.marks, self.value))

    def has_mark(self, mark: object) -> bool:
        return mark in self.marks

    def mark(self, mark: object) -> Self:
        return evolve(self, marks=self.marks.union({mark}))

    def with_marks(self, marks_to_add: set[Any]) -> Self:
        return evolve(self, marks=self.marks.union(marks_to_add))

    def unmark(self) -> tuple[Self, frozenset[Any]]:
        unmarked_value = evolve(self, marks=frozenset())
        return unmarked_value, self.marks

    def is_true(self) -> bool:
        from pyvider.cty.types import CtyDynamic

        if isinstance(self.vtype, CtyDynamic) and isinstance(self.value, CtyValue):
            return self.value.is_true()
        return self.value is True

    def is_false(self) -> bool:
        from pyvider.cty.types import CtyDynamic

        if isinstance(self.vtype, CtyDynamic) and isinstance(self.value, CtyValue):
            return self.value.is_false()
        return self.value is False

    def is_empty(self) -> bool:
        return not self.value if hasattr(self.value, "__len__") else False

    def with_key(self, key: str, value: Any) -> Self:
        from ..types import CtyMap

        if not isinstance(self.vtype, CtyMap):
            raise TypeError("'.with_key()' can only be used on CtyMap values.")
        if not isinstance(self.value, dict):
            raise TypeError("Internal value of CtyMap must be a dict.")
        new_dict = self.value.copy()
        new_dict[key] = value
        # validate() returns CtyValue[Any] due to .value: object limitation
        return self.vtype.validate(new_dict)  # type: ignore[no-any-return]

    def without_key(self, key: str) -> Self:
        from ..types import CtyMap

        if not isinstance(self.vtype, CtyMap):
            raise TypeError("'.without_key()' can only be used on CtyMap values.")
        if not isinstance(self.value, dict):
            raise TypeError("Internal value of CtyMap must be a dict.")
        if key not in self.value:
            return self
        new_dict = self.value.copy()
        del new_dict[key]
        # validate() returns CtyValue[Any] due to .value: object limitation
        return self.vtype.validate(new_dict)  # type: ignore[no-any-return]

    def append(self, value: Any) -> Self:
        from ..types import CtyList

        if not isinstance(self.vtype, CtyList):
            raise TypeError("'.append()' can only be used on CtyList values.")
        if not isinstance(self.value, list | tuple):
            raise TypeError("Internal value of CtyList must be a list or tuple.")
        new_list = list(self.value)
        new_list.append(value)
        # validate() returns CtyValue[Any] due to .value: object limitation
        return self.vtype.validate(new_list)  # type: ignore[no-any-return]

    def with_element_at(self, index: int, value: Any) -> Self:
        from ..types import CtyList

        if not isinstance(self.vtype, CtyList):
            raise TypeError("'.with_element_at()' can only be used on CtyList values.")
        if not isinstance(self.value, list | tuple):
            raise TypeError("Internal value of CtyList must be a list or tuple.")
        new_list = list(self.value)
        if not (-len(new_list) <= index < len(new_list)):
            raise IndexError("list index out of range")
        new_list[index] = value
        # validate() returns CtyValue[Any] due to .value: object limitation
        return self.vtype.validate(new_list)  # type: ignore[no-any-return]

    @classmethod
    def unknown(cls, vtype: CtyType[Any], value: Any = UNREFINED_UNKNOWN) -> CtyValue[Any]:
        return cls(vtype=vtype, is_unknown=True, value=value)

    @classmethod
    def null(cls, vtype: CtyType[Any]) -> CtyValue[Any]:
        return cls(vtype=vtype, is_null=True)


# 🌊🪢🔚



#
# SPDX-FileCopyrightText: Copyright (c) provide.io llc. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#


from __future__ import annotations

from decimal import Decimal
from typing import Any, cast

from pyvider.cty.types import (
    CtyDynamic,
    CtyList,
    CtyMap,
    CtyObject,
    CtySet,
    CtyTuple,
)

# Local imports to break the circular dependency cycle.
from pyvider.cty.values import CtyValue

# Module-level sentinel to avoid per-call allocation
_POST_PROCESS = object()


def cty_to_native(value: CtyValue[Any] | Any) -> Any:  # noqa: C901
    """
    Converts a CtyValue to its raw Python representation using an iterative
    approach to avoid recursion limits. This is safe for deeply nested structures.
    """
    if not isinstance(value, CtyValue):
        return value

    if value.is_unknown:
        return None  # Gracefully handle unknown values by returning None.

    if value.is_null:
        return None

    # Fast path for primitive types — avoid allocating work_stack/results/processing
    if not isinstance(
        value.type,
        CtyObject | CtyMap | CtyList | CtySet | CtyTuple | CtyDynamic,
    ):
        inner_val = value.value
        if isinstance(inner_val, Decimal):
            exponent = inner_val.as_tuple().exponent
            if isinstance(exponent, int) and exponent >= 0:
                return int(inner_val)
            return float(inner_val)
        return inner_val

    POST_PROCESS = _POST_PROCESS
    work_stack: list[Any] = [value]
    results: dict[int, Any] = {}
    processing: set[int] = set()

    while work_stack:
        current_item = work_stack.pop()

        if current_item is POST_PROCESS:
            val_to_process = work_stack.pop()
            val_id = id(val_to_process)
            processing.remove(val_id)

            # Robustness check for malformed collection values
            if not hasattr(val_to_process.value, "__iter__"):
                if isinstance(val_to_process.type, CtyList | CtySet):
                    results[val_id] = []
                elif isinstance(val_to_process.type, CtyTuple):
                    results[val_id] = ()
                elif isinstance(val_to_process.type, CtyMap | CtyObject):
                    results[val_id] = {}
                continue

            if isinstance(val_to_process.type, CtyDynamic):
                inner_id = id(val_to_process.value)
                results[val_id] = results[inner_id]
            elif isinstance(val_to_process.type, CtyObject | CtyMap):
                dict_val = cast(dict[str, Any], val_to_process.value)
                results[val_id] = {k: results[id(v)] for k, v in dict_val.items()}
            elif isinstance(val_to_process.type, CtyList):
                list_val = cast(list[Any], val_to_process.value)
                results[val_id] = [results[id(item)] for item in list_val]
            elif isinstance(val_to_process.type, CtySet):
                # Use _canonical_sort_key for consistent sorting of set elements
                set_val = cast(set[Any], val_to_process.value)
                results[val_id] = sorted(
                    [results[id(item)] for item in set_val],
                    key=lambda v: v._canonical_sort_key() if isinstance(v, CtyValue) else repr(v),
                )
            elif isinstance(val_to_process.type, CtyTuple):
                tuple_val = cast(tuple[Any, ...], val_to_process.value)
                results[val_id] = tuple(results[id(item)] for item in tuple_val)
            continue

        if not isinstance(current_item, CtyValue):
            results[id(current_item)] = current_item
            continue

        if current_item.is_unknown:
            results[id(current_item)] = None
            continue
        if current_item.is_null:
            results[id(current_item)] = None
            continue

        item_id = id(current_item)
        if item_id in results or item_id in processing:
            continue

        if isinstance(
            current_item.type,
            CtyObject | CtyMap | CtyList | CtySet | CtyTuple | CtyDynamic,
        ):
            processing.add(item_id)
            # Two appends avoid allocating a temporary 2-element list
            work_stack.append(current_item)
            work_stack.append(POST_PROCESS)

            if isinstance(current_item.type, CtyDynamic):
                work_stack.append(current_item.value)
            elif hasattr(current_item.value, "__iter__"):  # Robustness check
                if isinstance(current_item.value, dict):
                    # dict.values() supports reversed() in Python 3.8+, no list() needed
                    work_stack.extend(reversed(current_item.value.values()))
                elif isinstance(current_item.value, list | tuple):
                    # reversed() works directly on list/tuple, no intermediate list needed
                    work_stack.extend(reversed(current_item.value))
                else:
                    # sets need materialisation for reversed()
                    work_stack.extend(list(current_item.value))
        else:
            inner_val = current_item.value
            if isinstance(inner_val, Decimal):
                exponent = inner_val.as_tuple().exponent
                if isinstance(exponent, int) and exponent >= 0:
                    results[item_id] = int(inner_val)
                else:
                    results[item_id] = float(inner_val)
            else:
                results[item_id] = inner_val

    return results.get(id(value))


# 🌊🪢🔚



#
# SPDX-FileCopyrightText: Copyright (c) provide.io llc. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#


from __future__ import annotations

from pyvider.cty.context.validation_context import (
    MAX_VALIDATION_DEPTH,
    deeper_validation,
    get_validation_depth,
)

# pyvider/cty/context/__init__.py
"""
Provides context management for CTY operations.

This package includes tools for managing and retrieving the current operational
context within the CTY system, which can influence how types and values are
processed or validated.
"""

__all__ = [
    "MAX_VALIDATION_DEPTH",
    "deeper_validation",
    "get_validation_depth",
]

# 🌊🪢🔚



#
# SPDX-FileCopyrightText: Copyright (c) provide.io llc. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#

"""Comprehensive test suite for error context integration with provide-foundation.

This test suite validates that all error boundaries and enhanced error contexts
are working correctly across the pyvider-cty codebase."""

import pytest

from pyvider.cty import (
    CtyBool,
    CtyList,
    CtyMap,
    CtyNumber,
    CtyObject,
    CtyString,
    CtyValidationError,
)
from pyvider.cty.codec import cty_from_msgpack, cty_to_msgpack
from pyvider.cty.conversion import convert
from pyvider.cty.exceptions import (
    CtyAttributeValidationError,
    CtyFunctionError,
    CtyListValidationError,
    CtyMapValidationError,
    CtyStringValidationError,
)
from pyvider.cty.exceptions.encoding import DeserializationError
from pyvider.cty.functions.collection_functions import (
    concat,
    element,
    keys,
    length,
)
from pyvider.cty.parser import parse_tf_type_to_ctytype


class TestConversionErrorContext:
    """Test error context for type conversion operations."""

    def test_conversion_error_has_rich_context(self) -> None:
        """Test that conversion errors include rich context information."""
        string_val = CtyString().validate("not_a_number")

        with pytest.raises(Exception) as exc_info:
            convert(string_val, CtyNumber())

        # Should have conversion context
        error = exc_info.value
        assert hasattr(error, "context") or "conversion" in str(error).lower()

    def test_conversion_bool_error_context(self) -> None:
        """Test boolean conversion error context."""
        string_val = CtyString().validate("invalid_bool")

        with pytest.raises(Exception) as exc_info:
            convert(string_val, CtyBool())

        error = exc_info.value
        assert "bool" in str(error).lower()


class TestValidationErrorContext:
    """Test enhanced error context for validation operations."""

    def test_string_validation_error_context(self) -> None:
        """Test string validation with enhanced error context."""
        string_type = CtyString()

        class BadValue:
            def __str__(self) -> str:
                raise ValueError("Cannot convert to string")

        with pytest.raises(CtyStringValidationError) as exc_info:
            string_type.validate(BadValue())

        error = exc_info.value
        assert hasattr(error, "context") or "BadValue" in str(error)

    def test_list_validation_nested_error_context(self) -> None:
        """Test list validation with nested error context."""
        list_type = CtyList(element_type=CtyString())

        class BadElement:
            def __str__(self) -> str:
                raise ValueError("Bad element")

        with pytest.raises(CtyListValidationError) as exc_info:
            list_type.validate(["good", BadElement(), "also_good"])

        error = exc_info.value
        # Should indicate which element failed
        assert "[1]" in str(error) or "index 1" in str(error).lower()

    def test_map_validation_error_context(self) -> None:
        """Test map validation with key-specific error context."""
        map_type = CtyMap(element_type=CtyString())

        class BadValue:
            def __str__(self) -> str:
                raise ValueError("Bad map value")

        with pytest.raises(CtyMapValidationError) as exc_info:
            map_type.validate({"good": "value", "bad": BadValue()})

        error = exc_info.value
        # Should indicate which key failed
        assert "bad" in str(error).lower()

    def test_object_validation_error_context(self) -> None:
        """Test object validation with attribute-specific error context."""
        obj_type = CtyObject(attribute_types={"name": CtyString(), "age": CtyNumber()})

        with pytest.raises(CtyAttributeValidationError) as exc_info:
            obj_type.validate({"name": "Alice"})  # Missing required attribute

        error = exc_info.value
        assert "age" in str(error) or "Missing" in str(error)


class TestFunctionErrorContext:
    """Test error context for CTY standard library functions."""

    def test_length_function_error_context(self) -> None:
        """Test length function with invalid input type."""
        number_val = CtyNumber().validate(42)

        with pytest.raises(CtyFunctionError) as exc_info:
            length(number_val)

        error = exc_info.value
        assert "length" in str(error) and "number" in str(error).lower()

    def test_keys_function_error_context(self) -> None:
        """Test keys function with invalid input type."""
        string_val = CtyString().validate("not_a_map")

        with pytest.raises(CtyFunctionError) as exc_info:
            keys(string_val)

        error = exc_info.value
        assert "keys" in str(error) and "string" in str(error).lower()

    def test_element_function_error_context(self) -> None:
        """Test element function with invalid collection type."""
        string_val = CtyString().validate("not_a_list")
        index_val = CtyNumber().validate(0)

        with pytest.raises(CtyFunctionError) as exc_info:
            element(string_val, index_val)

        error = exc_info.value
        assert "element" in str(error) and "string" in str(error).lower()


class TestParserErrorContext:
    """Test error context for Terraform type parsing."""

    def test_invalid_primitive_type_error_context(self) -> None:
        """Test parsing invalid primitive type."""
        with pytest.raises(CtyValidationError) as exc_info:
            parse_tf_type_to_ctytype("invalid_type")

        error = exc_info.value
        assert "invalid_type" in str(error)

    def test_malformed_complex_type_error_context(self) -> None:
        """Test parsing malformed complex type specification."""
        with pytest.raises(CtyValidationError) as exc_info:
            parse_tf_type_to_ctytype(["list"])  # Missing element type

        error = exc_info.value
        assert "Invalid" in str(error) or "specification" in str(error)


class TestCodecErrorContext:
    """Test error context for serialization/deserialization."""

    def test_serialization_with_context(self) -> None:
        """Test that serialization operations include context."""
        # This should work normally, testing the error boundary is active
        string_val = CtyString().validate("test")
        data = cty_to_msgpack(string_val, CtyString())
        result = cty_from_msgpack(data, CtyString())
        assert result.raw_value == "test"

    def test_deserialization_invalid_data(self) -> None:
        """Test deserialization with invalid data."""
        with pytest.raises((DeserializationError, ValueError)):
            cty_from_msgpack(b"invalid_msgpack_data", CtyString())


class TestCompositeErrorContext:
    """Test error context in complex nested scenarios."""

    def test_complex_nested_validation_error_chain(self) -> None:
        """Test error context propagation through complex nested structures."""
        user_type = CtyObject(
            attribute_types={
                "profile": CtyObject(attribute_types={"settings": CtyMap(element_type=CtyString())})
            }
        )

        class BadValue:
            def __str__(self) -> str:
                raise ValueError("Bad nested value")

        with pytest.raises(CtyAttributeValidationError) as exc_info:
            user_type.validate({"profile": {"settings": {"theme": BadValue()}}})

        error = exc_info.value
        # Should show the full path to the error
        assert "profile" in str(error) or "settings" in str(error) or "theme" in str(error)

    def test_function_with_validation_error_chain(self) -> None:
        """Test function error context with underlying validation errors."""
        # Create a list with mixed valid/invalid elements for concat
        list1 = CtyList(element_type=CtyString()).validate(["valid"])
        list2 = CtyList(element_type=CtyString()).validate(["also_valid"])

        # This should work - testing error boundary presence
        result = concat(list1, list2)
        assert len(result.raw_value) == 2


class TestErrorRecoveryAndBoundaries:
    """Test that error boundaries don't interfere with normal operation."""

    def test_normal_operations_still_work(self) -> None:
        """Ensure error boundaries don't break normal functionality."""
        # Basic type validation
        string_val = CtyString().validate("hello")
        assert string_val.raw_value == "hello"

        # Collection operations
        list_val = CtyList(element_type=CtyString()).validate(["a", "b", "c"])
        length_val = length(list_val)
        assert length_val.raw_value == 3

        # Type conversion
        converted = convert(string_val, CtyString())  # Identity conversion
        assert converted.raw_value == "hello"

        # Type parsing
        parsed_type = parse_tf_type_to_ctytype("string")
        assert isinstance(parsed_type, CtyString)

    def test_error_boundaries_preserve_original_error_types(self) -> None:
        """Ensure error boundaries don't change the error types being raised."""
        with pytest.raises(CtyStringValidationError):
            CtyString().validate(123)  # Should still raise CtyStringValidationError

        with pytest.raises(CtyFunctionError):
            length(CtyNumber().validate(42))  # Should still raise CtyFunctionError


if __name__ == "__main__":
    pytest.main([__file__, "-v"])

# 🌊🪢🔚



#
# SPDX-FileCopyrightText: Copyright (c) provide.io llc. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#


from __future__ import annotations

from abc import ABC, abstractmethod
from typing import (
    TYPE_CHECKING,
    Any,
    ClassVar,
    Generic,
    Protocol,
    TypeVar,
    runtime_checkable,
)

from attrs import define

# Forward reference to CtyValue to avoid importing it directly at runtime
if TYPE_CHECKING:
    from pyvider.cty.values.base import CtyValue

T_co = TypeVar("T_co", covariant=True)
T = TypeVar("T")


@runtime_checkable
class CtyTypeProtocol(Protocol[T_co]):
    """Protocol defining the essential interface of a CtyType."""

    def validate(self, value: object) -> CtyValue[T_co]: ...
    def equal(self, other: Any) -> bool: ...
    def usable_as(self, other: Any) -> bool: ...
    def is_primitive_type(self) -> bool: ...


# The concrete ABC now implements the protocol
@define(slots=True)
class CtyType(CtyTypeProtocol[T], Generic[T], ABC):
    """
    Generic abstract base class for all Cty types.
    """

    ctype: ClassVar[str | None] = None
    _type_order: ClassVar[int] = 99

    @abstractmethod
    def validate(self, value: object) -> CtyValue[T]:
        pass

    @abstractmethod
    def equal(self, other: Any) -> bool:
        pass

    @abstractmethod
    def usable_as(self, other: Any) -> bool:
        pass

    @abstractmethod
    def _to_wire_json(self) -> Any:
        """Abstract method for JSON wire format encoding."""

    def is_primitive_type(self) -> bool:
        return False

    def is_dynamic_type(self) -> bool:
        """Returns True if this type is CtyDynamic."""
        return False

    def __eq__(self, other: object) -> bool:
        if isinstance(other, CtyType):
            return self.equal(other)
        return NotImplemented

    def __hash__(self) -> int:
        return hash(repr(self))

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}()"


# 🌊🪢🔚



#
# SPDX-FileCopyrightText: Copyright (c) provide.io llc. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#


from __future__ import annotations

from provide.foundation.utils.versioning import get_version

"""Version handling for pyvider-cty.

This module uses the shared versioning utility from provide-foundation.
"""

__version__ = get_version("pyvider-cty", caller_file=__file__)

__all__ = ["__version__"]

# 🌊🪢🔚



#
# SPDX-FileCopyrightText: Copyright (c) provide.io llc. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#


from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Any, TypeVar, cast

from attrs import define, field

from pyvider.cty.exceptions import (
    AttributePathError,
    CtyValidationError,
)
from pyvider.cty.types import CtyType
from pyvider.cty.values import CtyValue

T = TypeVar("T")


class PathStep(ABC):
    @abstractmethod
    def apply(self, value: CtyValue[Any]) -> CtyValue[Any]:
        pass

    @abstractmethod
    def apply_type(self, vtype: CtyType[Any]) -> CtyType[Any]:
        pass

    @abstractmethod
    def __str__(self) -> str:
        pass


@define(frozen=True)
class GetAttrStep(PathStep):
    name: str = field()

    @name.validator
    def _validate_name(self, attribute: str, value: str) -> None:
        if not value:
            raise ValueError("Attribute name cannot be empty")

    def apply(self, value: CtyValue[Any]) -> CtyValue[Any]:
        if value.is_null:
            raise AttributePathError(f"Cannot get attribute '{self.name}' from null value")
        from pyvider.cty.types.structural import CtyObject

        if isinstance(value.type, CtyObject):
            return value.type.get_attribute(value, self.name)
        raise AttributePathError(
            f"Cannot get attribute from non-object value of type {value.type.__class__.__name__}"
        )

    def apply_type(self, vtype: CtyType[Any]) -> CtyType[Any]:
        from pyvider.cty.types.structural import CtyObject

        if not isinstance(vtype, CtyObject):
            raise AttributePathError(f"Cannot get attribute from non-object type {vtype.__class__.__name__}")
        if not vtype.has_attribute(self.name):
            raise AttributePathError(f"Object type has no attribute {self.name}")
        return vtype.attribute_types[self.name]

    def __str__(self) -> str:
        return f".{self.name}"


@define(frozen=True)
class IndexStep(PathStep):
    index: int = field()

    def apply(self, value: CtyValue[Any]) -> CtyValue[Any]:
        if value.is_null:
            raise AttributePathError("Cannot index into null value")
        if value.is_unknown:
            return CtyValue.unknown(self.apply_type(value.type))
        from pyvider.cty.types.collections import CtyList
        from pyvider.cty.types.structural import CtyDynamic, CtyTuple

        if isinstance(value.type, CtyList | CtyTuple):
            list_or_tuple_type = cast(CtyList[Any] | CtyTuple, value.type)  # type: ignore[redundant-cast]
            return list_or_tuple_type.element_at(value, self.index)
        if isinstance(value.type, CtyDynamic) and isinstance(value.value, CtyValue):
            result = self.apply(value.value)
            return CtyValue(result.type, result.value)
        raise AttributePathError(f"Cannot index into value of type {type(value.type).__name__}")

    def apply_type(self, vtype: CtyType[Any]) -> CtyType[Any]:
        from pyvider.cty.types.collections import CtyList
        from pyvider.cty.types.structural import CtyDynamic, CtyTuple

        if isinstance(vtype, CtyList):
            return vtype.element_type
        if isinstance(vtype, CtyTuple):
            try:
                return vtype.element_types[self.index]
            except IndexError as e:
                raise AttributePathError(f"Tuple index {self.index} out of bounds") from e
        if isinstance(vtype, CtyDynamic):
            return CtyDynamic()
        raise AttributePathError(f"Cannot index into non-collection type {vtype.__class__.__name__}")

    def __str__(self) -> str:
        return f"[{self.index}]"


@define(frozen=True)
class KeyStep(PathStep):
    key: object = field()

    def apply(self, value: CtyValue[Any]) -> CtyValue[Any]:
        if value.is_null:
            raise AttributePathError("Cannot get key from null value")
        if value.is_unknown:
            return CtyValue.unknown(self.apply_type(value.type))
        from pyvider.cty.types.collections import CtyMap
        from pyvider.cty.types.structural import CtyDynamic

        if isinstance(value.type, CtyMap):
            return value.type.get(value, self.key)
        if isinstance(value.type, CtyDynamic) and isinstance(value.value, CtyValue):
            result = self.apply(value.value)
            return CtyValue(result.type, result.value)
        raise AttributePathError(
            f"Cannot get key from non-map/non-dynamic value of type {type(value.type).__name__}"
        )

    def apply_type(self, vtype: CtyType[Any]) -> CtyType[Any]:
        from pyvider.cty.types import CtyString
        from pyvider.cty.types.collections import CtyMap
        from pyvider.cty.types.structural import CtyDynamic

        if isinstance(vtype, CtyDynamic):
            return CtyDynamic()
        if not isinstance(vtype, CtyMap):
            raise AttributePathError(f"Cannot get key from non-map type {vtype.__class__.__name__}")
        try:
            CtyString().validate(self.key)
        except CtyValidationError as e:
            raise AttributePathError(f"Invalid key for map: {self.key!r} is not a valid string") from e
        return vtype.element_type

    def __str__(self) -> str:
        return f"[{self.key!r}]"


@define
class CtyPath:
    steps: list[PathStep] = field(factory=list)

    @classmethod
    def empty(cls) -> CtyPath:
        return cls([])

    @classmethod
    def get_attr(cls, name: str) -> CtyPath:
        return cls([GetAttrStep(name)])

    @classmethod
    def index(cls, index: int) -> CtyPath:
        return cls([IndexStep(index)])

    @classmethod
    def key(cls, key: object) -> CtyPath:
        return cls([KeyStep(key)])

    def child(self, name: str) -> CtyPath:
        return CtyPath([*self.steps, GetAttrStep(name)])

    def index_step(self, index: int) -> CtyPath:
        return CtyPath([*self.steps, IndexStep(index)])

    def key_step(self, key: object) -> CtyPath:
        return CtyPath([*self.steps, KeyStep(key)])

    def apply_path(self, value: object) -> CtyValue[Any]:
        if not self.steps:
            if isinstance(value, CtyValue):
                return value
            raise AttributePathError("Cannot return non-CtyValue from apply_path")
        if not isinstance(value, CtyValue):
            raise AttributePathError(f"Cannot apply path to non-CtyValue: {type(value).__name__}")
        current = value
        for i, step in enumerate(self.steps):
            try:
                current = step.apply(current)
            except AttributePathError as e:
                raise AttributePathError(f"Error at step {i + 1} ({step}): {e}") from e
        return current

    def apply_path_type(self, vtype: CtyType[Any]) -> CtyType[Any]:
        if not self.steps:
            return vtype
        current = vtype
        for i, step in enumerate(self.steps):
            try:
                current = step.apply_type(current)
            except AttributePathError as e:
                raise AttributePathError(f"Error at type step {i + 1} ({step}): {e}") from e
        return current

    def string(self) -> str:
        if not self.steps:
            return "(root)"

        path_str = ""
        for i, step in enumerate(self.steps):
            current_step_str = str(step)
            if i == 0 and isinstance(step, GetAttrStep):
                path_str += current_step_str[1:]  # Strip leading dot
            else:
                path_str += current_step_str

        return path_str

    def __str__(self) -> str:
        return self.string()


# 🌊🪢🔚



#
# SPDX-FileCopyrightText: Copyright (c) provide.io llc. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#


from __future__ import annotations

from pyvider.cty.values.base import CtyValue
from pyvider.cty.values.markers import (
    UNREFINED_UNKNOWN,
    RefinedUnknownValue,
    UnknownValue,
)

#
# pyvider/cty/values/__init__.py
#
"""
CTY Value Representation.

This package defines CtyValue, the runtime representation of values
within the CTY type system. CtyValue instances pair a Python value
with its corresponding CtyType and associated metadata.
"""

__all__ = [
    "UNREFINED_UNKNOWN",
    "CtyValue",
    "RefinedUnknownValue",
    "UnknownValue",
]

# 🌊🪢🔚



#
# SPDX-FileCopyrightText: Copyright (c) provide.io llc. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#

"""Advanced recursion detection for CTY validation.

This module provides sophisticated recursion detection that can distinguish between:
1. Genuine circular references that would cause infinite loops
2. Normal nested data structures with repetitive patterns
3. Deep but finite nesting that should be allowed

The implementation is designed for production IaC requirements where:
- Complex configurations with deep nesting must be supported
- Genuine circular references must be prevented
- Performance must be predictable and measurable
- Debugging and monitoring capabilities are essential"""

from __future__ import annotations

from collections.abc import Callable
from dataclasses import dataclass, field
from functools import wraps
import threading
import time
from typing import Any, cast

from provide.foundation import logger

from pyvider.cty.config.defaults import (
    MAX_OBJECT_REVISITS,
    MAX_VALIDATION_DEPTH,
    MAX_VALIDATION_TIME_MS,
)


@dataclass
class RecursionContext:
    """Thread-local context for tracking validation recursion."""

    # Object identity tracking: maps object id -> visit count (lightweight int counter)
    validation_graph: dict[int, int] = field(default_factory=dict)

    # Path tracking for detailed diagnostics (None entries are lightweight depth markers)
    validation_path: list[str | None] = field(default_factory=list)

    # Performance monitoring
    max_depth_reached: int = 0
    total_validations: int = 0
    validation_start_time: float = field(default_factory=time.time)

    # Configuration thresholds
    max_depth_allowed: int = MAX_VALIDATION_DEPTH
    max_object_revisits: int = MAX_OBJECT_REVISITS
    max_validation_time_ms: int = MAX_VALIDATION_TIME_MS

    # Flag to indicate validation was stopped due to recursion detection
    validation_stopped: bool = False

    def reset(self) -> None:
        """Reset context for new validation session."""
        self.validation_graph.clear()
        self.validation_path.clear()
        self.max_depth_reached = 0
        self.total_validations = 0
        self.validation_start_time = time.time()
        self.validation_stopped = False


# Thread-local storage for recursion contexts
_thread_local = threading.local()


def get_recursion_context() -> RecursionContext:
    """Get or create thread-local recursion context."""
    if not hasattr(_thread_local, "recursion_context"):
        _thread_local.recursion_context = RecursionContext()
    return cast(RecursionContext, _thread_local.recursion_context)


def clear_recursion_context() -> None:
    """Clear thread-local recursion context."""
    if hasattr(_thread_local, "recursion_context"):
        _thread_local.recursion_context.reset()


class RecursionDetector:
    """
    Advanced recursion detector for CTY validation.

    This detector uses sophisticated algorithms to distinguish between:
    - Circular references (object A -> object B -> object A)
    - Deep but finite nesting (legitimate complex configurations)
    - Performance pathological cases (excessive validation time)
    """

    def __init__(self, context: RecursionContext | None = None) -> None:
        self.context = context or get_recursion_context()

    def should_continue_validation(self, value: Any, current_path: str = "", /) -> tuple[bool, str | None]:
        """
        Determine if validation should continue for the given value.

        Returns:
            (should_continue, reason_if_stopped)

        Production requirements:
        - Must handle legitimate deep nesting (1000+ levels)
        - Must detect genuine circular references quickly
        - Must provide detailed diagnostics for debugging
        - Must have predictable performance characteristics
        """

        # Performance safeguards - prevent pathological cases
        # Only check time every 64 validations to reduce time.time() overhead
        if self.context.total_validations & 63 == 0:
            elapsed_ms = (time.time() - self.context.validation_start_time) * 1000
        else:
            elapsed_ms = 0.0
        if elapsed_ms > self.context.max_validation_time_ms:
            reason = (
                f"Validation timeout after {elapsed_ms:.1f}ms (max: {self.context.max_validation_time_ms}ms)"
            )
            logger.warning(
                "CTY validation timeout exceeded",
                elapsed_ms=elapsed_ms,
                max_allowed_ms=self.context.max_validation_time_ms,
                path=current_path,
                trace="advanced_recursion_detection",
            )
            return False, reason

        # Update context
        self.context.total_validations += 1
        current_depth = len(self.context.validation_path)
        self.context.max_depth_reached = max(self.context.max_depth_reached, current_depth)

        # Depth safeguards - only trigger for truly deep recursion
        if current_depth > self.context.max_depth_allowed:
            reason = f"Maximum nesting depth exceeded: {current_depth} > {self.context.max_depth_allowed}"
            logger.warning(
                "CTY validation depth limit exceeded",
                current_depth=current_depth,
                max_allowed=self.context.max_depth_allowed,
                path=current_path,
                trace="advanced_recursion_detection",
            )
            return False, reason

        # Skip cycle detection for primitive types and simple collections (performance optimization)
        if isinstance(value, (str, int, float, bool, type(None))):
            return True, None

        # Skip cycle detection for simple lists/tuples of primitives
        if isinstance(value, (list, tuple)) and all(
            isinstance(item, (str, int, float, bool, type(None))) for item in value
        ):
            return True, None

        # Lightweight cycle detection using visit counters
        value_id = id(value)
        visits = self.context.validation_graph.get(value_id, 0) + 1
        self.context.validation_graph[value_id] = visits

        if visits > self.context.max_object_revisits:
            value_type = type(value).__name__
            reason = (
                f"Circular reference detected: {value_type} object visited "
                f"{visits} times (max: {self.context.max_object_revisits})"
            )
            logger.debug(
                "CTY circular reference detected",
                object_type=value_type,
                object_id=value_id,
                visits=visits,
                current_depth=current_depth,
                path=current_path,
                trace="advanced_recursion_detection",
            )
            return False, reason

        return True, None

    def enter_validation_scope(self, scope_name: str) -> None:
        """Enter a new validation scope for path tracking."""
        self.context.validation_path.append(scope_name)

    def exit_validation_scope(self) -> None:
        """Exit the current validation scope."""
        if self.context.validation_path:
            self.context.validation_path.pop()

    def get_current_path(self) -> str:
        """Get the current validation path for diagnostics."""
        return " -> ".join(s for s in self.context.validation_path if s is not None)

    def get_performance_metrics(self) -> dict[str, Any]:
        """Get performance metrics for monitoring and debugging."""
        elapsed_ms = (time.time() - self.context.validation_start_time) * 1000
        return {
            "total_validations": self.context.total_validations,
            "max_depth_reached": self.context.max_depth_reached,
            "elapsed_ms": elapsed_ms,
            "objects_in_graph": len(self.context.validation_graph),
            "avg_validations_per_ms": self.context.total_validations / max(elapsed_ms, 0.001),
            "current_path": self.get_current_path(),
        }


def with_recursion_detection(func: Callable[..., Any]) -> Callable[..., Any]:
    """
    Decorator for advanced recursion detection in validation functions.
    """
    # Pre-allocate a single detector instance per decorated function.
    # The detector is stateless — context comes from thread-local storage.
    _detector = RecursionDetector()

    @wraps(func)
    def wrapper(self: Any, value: Any, *args: Any, **kwargs: Any) -> Any:
        context = get_recursion_context()
        # A call is top-level if the context has not been used yet.
        is_top_level_call = context.total_validations == 0
        if is_top_level_call:
            context.reset()

        # Bind detector to current thread's context (avoids per-call allocation)
        _detector.context = context

        # Use None as a lightweight depth marker instead of an f-string scope name.
        # The actual scope string is only constructed on the error path.
        context.validation_path.append(None)

        try:
            # Check if validation was already stopped by a nested call
            if context.validation_stopped:
                from pyvider.cty.values import CtyValue

                return CtyValue.unknown(self)

            should_continue, reason = _detector.should_continue_validation(value)
            if not should_continue:
                from pyvider.cty.values import CtyValue

                # Set flag to stop all parent validations
                context.validation_stopped = True

                # Only construct debug strings on the error path
                scope_name = f"{self.__class__.__name__}.validate(type={type(value).__name__})"
                logger.warning(
                    "CTY validation stopped due to recursion detection",
                    reason=reason,
                    value_type=type(value).__name__,
                    path=scope_name,
                )
                return CtyValue.unknown(self)

            # The decorator no longer passes the internal flag down.
            result = func(self, value, *args, **kwargs)

            # Check again after validation in case a nested call stopped validation
            if context.validation_stopped:
                from pyvider.cty.values import CtyValue

                return CtyValue.unknown(self)

            return result
        finally:
            if context.validation_path:
                context.validation_path.pop()

    return wrapper


# 🌊🪢🔚



#
# SPDX-FileCopyrightText: Copyright (c) provide.io llc. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#


import pytest

from pyvider.cty.exceptions import CtyValidationError
from pyvider.cty.parser import parse_tf_type_to_ctytype
from pyvider.cty.types import (
    CtyBool,
    CtyDynamic,
    CtyList,
    CtyMap,
    CtyNumber,
    CtyObject,
    CtySet,
    CtyString,
    CtyTuple,
)

valid_test_cases = [
    ("string", CtyString()),
    ("number", CtyNumber()),
    ("bool", CtyBool()),
    ("dynamic", CtyDynamic()),
    (["list", "string"], CtyList(element_type=CtyString())),
    (["set", "number"], CtySet(element_type=CtyNumber())),
    (["map", "bool"], CtyMap(element_type=CtyBool())),
    (
        ["object", {"name": "string", "enabled": "bool"}],
        CtyObject(attribute_types={"name": CtyString(), "enabled": CtyBool()}),
    ),
    (
        ["tuple", ["string", "number"]],
        CtyTuple(element_types=(CtyString(), CtyNumber())),
    ),
]

invalid_test_cases = [
    "taco",
    ["list"],
    ["list", "taco"],
    ["object", ["name", "string"]],
    ["tuple", {"name": "string"}],
    ["taco", "string"],
]


@pytest.mark.parametrize("type_spec, expected_type", valid_test_cases)
def test_valid_type_parsing(type_spec, expected_type) -> None:
    result = parse_tf_type_to_ctytype(type_spec)
    assert result.equal(expected_type)


@pytest.mark.parametrize("type_spec", invalid_test_cases)
def test_invalid_type_parsing_raises_error(type_spec) -> None:
    with pytest.raises(CtyValidationError):
        parse_tf_type_to_ctytype(type_spec)


# 🌊🪢🔚



#
# SPDX-FileCopyrightText: Copyright (c) provide.io llc. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#


from __future__ import annotations

from pyvider.cty.path.base import (
    CtyPath,
    GetAttrStep,
    IndexStep,
    KeyStep,
    PathStep,
)

#
# pyvider/cty/path/__init__.py
#
"""
Provides CTY path navigation capabilities.

This package defines classes and utilities for constructing and applying
paths to navigate through nested CTY data structures (objects, lists, maps, tuples),
similar to property accessors or indexers in other languages.
"""

__all__ = [
    "CtyPath",
    "GetAttrStep",
    "IndexStep",
    "KeyStep",
    "PathStep",
]

# 🌊🪢🔚



#
# SPDX-FileCopyrightText: Copyright (c) provide.io llc. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#


from __future__ import annotations

from typing import TYPE_CHECKING, Any, ClassVar
import unicodedata

from attrs import define

from pyvider.cty.exceptions import CtyStringValidationError
from pyvider.cty.types.base import CtyType

if TYPE_CHECKING:
    from pyvider.cty.values import CtyValue


@define(frozen=True, slots=True)
class CtyString(CtyType[str]):
    ctype: ClassVar[str] = "string"
    _type_order: ClassVar[int] = 1

    def validate(self, value: object) -> CtyValue[str]:
        from pyvider.cty.values import CtyValue, UnknownValue

        if isinstance(value, UnknownValue):
            return CtyValue.unknown(self)

        if isinstance(value, CtyValue):
            if value.is_null:
                return CtyValue.null(self)
            if value.is_unknown:
                return CtyValue.unknown(self)
            raw_value = value.value
        else:
            raw_value = value

        if raw_value is None:
            return CtyValue.null(self)

        if not isinstance(raw_value, str | bytes):
            raise CtyStringValidationError(f"Cannot convert {type(raw_value).__name__} to string.")

        try:
            str_value = raw_value.decode("utf-8") if isinstance(raw_value, bytes) else str(raw_value)
            normalized_value = unicodedata.normalize("NFC", str_value)
            return CtyValue(vtype=self, value=normalized_value)
        except Exception as e:
            raise CtyStringValidationError(f"Cannot convert {type(raw_value).__name__} to string: {e}") from e

    def equal(self, other: CtyType[Any]) -> bool:
        return isinstance(other, CtyString)

    def usable_as(self, other: CtyType[Any]) -> bool:
        from pyvider.cty.types.structural import CtyDynamic

        return isinstance(other, CtyString | CtyDynamic)

    def _to_wire_json(self) -> Any:
        return self.ctype

    def __str__(self) -> str:
        return "string"

    def is_primitive_type(self) -> bool:
        return True


# 🌊🪢🔚



#
# SPDX-FileCopyrightText: Copyright (c) provide.io llc. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#


from __future__ import annotations

from pyvider.cty.types.base import CtyType
from pyvider.cty.types.capsule import CtyCapsule, CtyCapsuleWithOps
from pyvider.cty.types.capsule_types import BytesCapsule
from pyvider.cty.types.collections import (
    CtyList,
    CtyMap,
    CtySet,
)
from pyvider.cty.types.primitives import (
    CtyBool,
    CtyNumber,
    CtyString,
)
from pyvider.cty.types.structural import (
    CtyDynamic,
    CtyObject,
    CtyTuple,
)

#
# pyvider/cty/types/__init__.py
#
"""
Defines the core CTY (Compatible Type System) types.

This package includes the base CtyType class and all concrete type
implementations such as primitives (string, number, bool), collections
(list, map, set), and structural types (object, tuple, dynamic, capsule).
"""

__all__ = [
    "BytesCapsule",
    "CtyBool",
    "CtyCapsule",
    "CtyCapsuleWithOps",
    "CtyDynamic",
    "CtyList",
    "CtyMap",
    "CtyNumber",
    "CtyObject",
    "CtySet",
    "CtyString",
    "CtyTuple",
    "CtyType",
]

# 🌊🪢🔚



#
# SPDX-FileCopyrightText: Copyright (c) provide.io llc. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#


from __future__ import annotations

from pyvider.cty.types.collections.list import CtyList
from pyvider.cty.types.collections.map import CtyMap
from pyvider.cty.types.collections.set import CtySet

#
# pyvider/cty/types/collections/__init__.py
#
"""
CTY Collection Types.

This package implements the collection types for the CTY system,
including List, Map, and Set types. These types manage collections
of CTY values with specific constraints (e.g., element types, uniqueness).
"""

__all__ = [
    "CtyList",
    "CtyMap",
    "CtySet",
]

# 🌊🪢🔚



#
# SPDX-FileCopyrightText: Copyright (c) provide.io llc. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#


from __future__ import annotations

"""Centralized default values for pyvider-cty configuration.
All defaults are defined here instead of inline in field definitions.
"""

# =================================
# Performance and caching defaults
# =================================
ENABLE_TYPE_INFERENCE_CACHE = True  # Enable caching for type inference performance

# =================================
# Validation defaults
# =================================
MAX_VALIDATION_DEPTH = 500  # Safer default, well below Python's typical limit
MAX_OBJECT_REVISITS = 100  # Allow many revisits for complex schemas
MAX_VALIDATION_TIME_MS = 30000  # 30 second timeout for pathological cases

# =================================
# Codec defaults
# =================================
MSGPACK_EXT_TYPE_CTY = 0
MSGPACK_EXT_TYPE_REFINED_UNKNOWN = 12
MSGPACK_RAW_FALSE = False
MSGPACK_STRICT_MAP_KEY_FALSE = False
MSGPACK_USE_BIN_TYPE_TRUE = True

# Refinement payload field IDs
REFINEMENT_IS_KNOWN_NULL = 1
REFINEMENT_STRING_PREFIX = 2
REFINEMENT_NUMBER_LOWER_BOUND = 3
REFINEMENT_NUMBER_UPPER_BOUND = 4
REFINEMENT_COLLECTION_LENGTH_LOWER_BOUND = 5
REFINEMENT_COLLECTION_LENGTH_UPPER_BOUND = 6

# =================================
# Function operation constants
# =================================
NUMERIC_OPERATIONS = frozenset(["add", "subtract", "multiply", "divide"])
COMPARISON_OPERATIONS = frozenset(["max", "min"])
TIME_UNITS = frozenset(["h", "m", "s"])

# =================================
# Collection defaults
# =================================
EMPTY_LIST_SIZE = 0
NEGATIVE_ONE_LENGTH = -1  # Used for "rest of string" operations

# =================================
# Comparison defaults
# =================================
COMPARISON_OPS_MAP = {
    ">": lambda x, y: x > y,
    ">=": lambda x, y: x >= y,
    "<": lambda x, y: x < y,
    "<=": lambda x, y: x <= y,
}

# =================================
# Parser type kinds
# =================================
TYPE_KIND_LIST = "list"
TYPE_KIND_SET = "set"
TYPE_KIND_MAP = "map"
TYPE_KIND_OBJECT = "object"
TYPE_KIND_TUPLE = "tuple"

# =================================
# Time conversion constants
# =================================
SECONDS_PER_HOUR = 3600
SECONDS_PER_MINUTE = 60
SECONDS_PER_SECOND = 1

# =================================
# Zero/null/boundary values
# =================================
ZERO_VALUE = 0
POSITIVE_BOUNDARY = 0
ONE_VALUE = 1
TWO_VALUE = 2

# =================================
# Common thresholds and limits
# =================================
DEFAULT_MAX_ITERATIONS = 1000
DEFAULT_TIMEOUT_SECONDS = 30
MAX_STRING_LENGTH_DISPLAY = 100
MAX_RECURSION_DEPTH = 100
MIN_COLLECTION_SIZE = 0

# =================================
# Exception message templates
# =================================
# These are used to avoid raw strings in exceptions
ERR_DECODE_REFINED_UNKNOWN = "Failed to decode refined unknown payload: {error}"
ERR_DYNAMIC_MALFORMED = "CtyDynamic value is malformed; its inner value is not a CtyValue instance."
ERR_DECODE_DYNAMIC_TYPE = "Failed to decode dynamic value type spec from JSON"
ERR_VALUE_FOR_OBJECT = "Value for CtyObject must be a dict"
ERR_VALUE_FOR_MAP = "Value for CtyMap must be a dict"
ERR_VALUE_FOR_LIST_SET = "Value for CtyList or CtySet must be iterable"
ERR_VALUE_FOR_TUPLE = "Value for CtyTuple must be a tuple"
ERR_OBJECT_NOT_MSGPACK_SERIALIZABLE = "Object of type {type_name} is not MessagePack serializable"
ERR_CANNOT_COMPARE = "Cannot compare {type1} with {type2}"
ERR_ALL_ARGS_SAME_TYPE = "All arguments to {op} must be of the same type (all numbers or all strings)"
ERR_MIN_ONE_ARG = "{op} requires at least one argument"
ERR_CANNOT_INFER_FROM_CTY_TYPE = "Cannot infer data type from a CtyType instance: {type_name}"
ERR_CANNOT_INFER_FROM_CTY_VALUE = "Cannot infer data type from a CtyValue instance: {type_name}"
ERR_VALUE_TYPE_NO_LEN = "Values of type {type_name} do not have a length"
ERR_MISSING_REQUIRED_ATTRIBUTE = "Missing required attribute '{name}'"

# Conversion error messages
ERR_CAPSULE_CANNOT_CONVERT = "Capsule type {value_type} cannot be converted to {target_type}"
ERR_CUSTOM_CONVERTER_NON_CTYVALUE = "Custom capsule converter returned a non-CtyValue object"
ERR_CUSTOM_CONVERTER_WRONG_TYPE = (
    "Custom capsule converter returned a value of the wrong type (got {result_type}, want {target_type})"
)
ERR_DYNAMIC_VALUE_NOT_CTYVALUE = "Dynamic value does not contain a CtyValue"
ERR_CANNOT_CONVERT_VALIDATION = "Cannot convert {value_type} to {target_type}: {message}"
ERR_CANNOT_CONVERT_TO_BOOL = "Cannot convert {value_type} to bool"
ERR_SOURCE_OBJECT_NOT_DICT = "Source object is not a dictionary"
ERR_CANNOT_CONVERT_GENERAL = "Cannot convert from {value_type} to {target_type}"

# Type system error messages
ERR_CONVERT_FN_MUST_BE_CALLABLE = "`convert_fn` must be a callable that accepts 2 arguments"
ERR_EQUAL_FN_MUST_BE_CALLABLE = "`equal_fn` must be a callable that accepts 2 arguments"
ERR_HASH_FN_MUST_BE_CALLABLE = "`hash_fn` must be a callable that accepts 1 argument"
ERR_ELEMENT_TYPE_MUST_BE_CTYTYPE = "element_type must be a CtyType instance, got {type_name}"
ERR_ELEMENT_TYPES_MUST_BE_TUPLE = "element_types must be a tuple, got {type_name}"
ERR_ELEMENT_TYPE_AT_INDEX_MUST_BE_CTYTYPE = "Element type at index {index} must be a CtyType, got {type_name}"
ERR_EXPECTED_CTYTYPE = "Expected CtyType, but got {type_name}"
ERR_EXPECTED_CTYTYPE_FOR_ELEMENT = "Expected CtyType for element_type, got {type_name}"
ERR_OBJECT_TYPE_SPEC_MUST_BE_DICT = "Object type spec must be a dictionary, got {type_name}"
ERR_TUPLE_TYPE_SPEC_MUST_BE_LIST = "Tuple type spec must be a list, got {type_name}"
ERR_ATTRIBUTE_MUST_BE_CTYTYPE = "Attribute '{name}' must be a CtyType, but got {type_name}"
ERR_INVALID_TERRAFORM_TYPE = "Invalid Terraform type specification: {tf_type}"
ERR_UNKNOWN_PRIMITIVE_TYPE = "Unknown primitive type name: '{tf_type}'"

# Value validation error messages
ERR_EXPECTED_DICT_FOR_OBJECT = "Expected a dictionary for CtyObject, got {type_name}"
ERR_EXPECTED_LIST_TUPLE_FOR_TUPLE = "Expected tuple or list, got {type_name}"
ERR_EXPECTED_SET_LIST_TUPLE = "Expected a Python set, frozenset, list, or tuple, got {type_name}"
ERR_EXPECTED_ELEMENTS_COUNT = "Expected {expected} elements, got {actual}"
ERR_OBJECT_ATTRIBUTE_NAME_MUST_BE_STRING = "Object attribute name must be a string, got {type_name}"
ERR_MAP_KEYS_MUST_BE_STRINGS = "Map keys must be strings, but got key of type {type_name}"
ERR_INVALID_MAP_KEY = "Invalid key for map: {key} is not a valid string"
ERR_VALUE_NOT_LIST_TUPLE = "Value to validate is not a list or tuple, but {type_name}"
ERR_INPUT_MUST_BE_DICT = "Input must be a dictionary, got {type_name}."
ERR_VALUE_NOT_INSTANCE = "Value is not an instance of {expected_type}. Got {actual_type}."
ERR_CTYLIST_VALUE_NOT_LIST_TUPLE = "CtyList value is not a list/tuple, but {type_name}"
ERR_LIST_ELEMENTS_CANNOT_BE_NULL = "List elements cannot be null for element type {element_type}"
ERR_CTYOBJECT_VALUE_NOT_DICT = "CtyObject value is not a dict"
ERR_CONTAINER_VALUE_MUST_BE_CTYVALUE = (
    "Container value must be a CtyValue of type {expected_type}, got {actual_value}"
)

# Path and navigation error messages
ERR_CANNOT_APPLY_PATH_NON_CTYVALUE = "Cannot apply path to non-CtyValue: {type_name}"
ERR_CANNOT_RETURN_NON_CTYVALUE_FROM_APPLY_PATH = "Cannot return non-CtyValue from apply_path"
ERR_TUPLE_INDEX_OUT_OF_BOUNDS = "Tuple index {index} out of bounds"
ERR_OBJECT_NO_ATTRIBUTE = "Object has no attribute '{name}'"
ERR_OBJECT_TYPE_NO_ATTRIBUTE = "Object type has no attribute {name}"

# Collection operation error messages
ERR_CANNOT_INDEX_NON_COLLECTION_TYPE = "Cannot index into non-collection type {type_name}"
ERR_CANNOT_INDEX_NON_COLLECTION_VALUE = "Cannot index into value of type {type_name}"
ERR_CANNOT_GET_ATTRIBUTE_NON_OBJECT_TYPE = "Cannot get attribute from non-object type {type_name}"
ERR_CANNOT_GET_ATTRIBUTE_NON_OBJECT_VALUE = "Cannot get attribute from non-object value of type {type_name}"
ERR_CANNOT_GET_KEY_NON_MAP_TYPE = "Cannot get key from non-map type {type_name}"
ERR_CANNOT_GET_KEY_NON_MAP_VALUE = "Cannot get key from non-map/non-dynamic value of type {type_name}"
ERR_LIST_INDICES_MUST_BE_INT = "list indices must be integers or slices, not {type_name}"
ERR_TUPLE_INDICES_MUST_BE_INT = "Tuple indices must be integers or slices, not {type_name}"

# Value access error messages
ERR_CANNOT_ACCESS_ELEMENT_NULL_LIST = "Cannot access element at index {index} in a null list."
ERR_CANNOT_GET_ATTRIBUTE_NULL_VALUE = "Cannot get attribute '{name}' from null value"
ERR_CANNOT_INDEX_NULL_VALUE = "Cannot index into null value"
ERR_CANNOT_INDEX_UNKNOWN_NULL_VALUE = "Cannot index into unknown or null value"
ERR_CANNOT_GET_KEY_NULL_VALUE = "Cannot get key from null value"
ERR_CANNOT_ITERATE_UNKNOWN_VALUE = "Cannot iterate unknown value"
ERR_CANNOT_GET_LENGTH_UNKNOWN_VALUE = "Cannot get length of unknown value"
ERR_CANNOT_GET_RAW_VALUE_UNKNOWN = "Cannot get raw value of unknown value"
ERR_CANNOT_COMPARE_NULL_UNKNOWN = "Cannot compare null or unknown values"

# CtyValue method error messages
ERR_APPEND_ONLY_FOR_CTYLIST = "'.append()' can only be used on CtyList values."
ERR_WITH_ELEMENT_AT_ONLY_FOR_CTYLIST = "'.with_element_at()' can only be used on CtyList values."
ERR_WITH_KEY_ONLY_FOR_CTYMAP = "'.with_key()' can only be used on CtyMap values."
ERR_WITHOUT_KEY_ONLY_FOR_CTYMAP = "'.without_key()' can only be used on CtyMap values."
ERR_INTERNAL_VALUE_CTYLIST_MUST_BE_LIST_TUPLE = "Internal value of CtyList must be a list or tuple."
ERR_INTERNAL_VALUE_CTYMAP_MUST_BE_DICT = "Internal value of CtyMap must be a dict."
ERR_LIST_INDEX_OUT_OF_RANGE = "list index out of range"
ERR_TUPLE_INDEX_OUT_OF_RANGE = "tuple index out of range"
ERR_INTERNAL_TUPLE_VALUE_INCONSISTENT = "Internal tuple value is inconsistent with type definition."
ERR_INTERNAL_TUPLE_VALUE_INCONSISTENT_SLICING = (
    "Internal tuple value is inconsistent with type definition for slicing."
)

# Comparison and type error messages
ERR_CANNOT_COMPARE_CTYVALUE_WITH = "Cannot compare CtyValue with {type_name}"
ERR_CANNOT_COMPARE_DIFFERENT_TYPES = "Cannot compare CtyValues of different types: {type1} and {type2}"
ERR_VALUE_TYPE_NOT_COMPARABLE = "Value of type {type} is not comparable"
ERR_VALUE_TYPE_NO_LEN = "Value of type {type_name} has no len()"
ERR_VALUE_TYPE_NOT_ITERABLE = "Value of type {type_name} is not iterable"
ERR_VALUE_TYPE_NOT_SUBSCRIPTABLE = "Value of type {type_name} is not subscriptable"
ERR_UNHASHABLE_TYPE = "unhashable type: 'CtyValue[{vtype}]'"

# Conversion type error messages
ERR_CANNOT_CONVERT_BOOL = "Cannot convert {type_name} to bool."
ERR_CANNOT_CONVERT_STRING = "Cannot convert {type_name} to string."
ERR_CANNOT_CONVERT_STRING_WITH_ERROR = "Cannot convert {type_name} to string: {error}"
ERR_CANNOT_REPRESENT_AS_DECIMAL = "Cannot represent {type_name} value '{raw_value}' as Decimal"

# Function-specific error messages
ERR_ADD_ARGS_MUST_BE_NUMBERS = "add: arguments must be numbers"
ERR_SUBTRACT_ARGS_MUST_BE_NUMBERS = "subtract: arguments must be numbers"
ERR_MULTIPLY_ARGS_MUST_BE_NUMBERS = "multiply: arguments must be numbers"
ERR_DIVIDE_ARGS_MUST_BE_NUMBERS = "divide: arguments must be numbers"
ERR_DIVIDE_BY_ZERO = "divide by zero"
ERR_MODULO_ARGS_MUST_BE_NUMBERS = "modulo: arguments must be numbers"
ERR_MODULO_BY_ZERO = "modulo by zero"
ERR_NEGATE_ARG_MUST_BE_NUMBER = "negate: argument must be a number"
ERR_POW_ARGS_MUST_BE_NUMBERS = "pow: arguments must be numbers"
ERR_POW_INVALID_OPERATION = "pow: invalid operation: {error}"
ERR_ABS_MUST_BE_NUMBER = "abs: input must be a number, got {type}"
ERR_CEIL_MUST_BE_NUMBER = "ceil: input must be a number, got {type}"
ERR_FLOOR_MUST_BE_NUMBER = "floor: input must be a number, got {type}"
ERR_SIGNUM_MUST_BE_NUMBER = "signum: input must be a number, got {type}"
ERR_INT_MUST_BE_NUMBER = "int: argument must be a number, got {type}"
ERR_LOG_ARGS_MUST_BE_NUMBERS = "log: arguments must be numbers"
ERR_LOG_BASE_CANNOT_BE_ONE = "log: base cannot be 1"
ERR_LOG_BASE_MUST_BE_POSITIVE = "log: base must be positive, got {base}"
ERR_LOG_NUMBER_MUST_BE_POSITIVE = "log: number must be positive, got {number}"
ERR_LOG_MATH_DOMAIN_ERROR = "log: math domain error: {error}"

# String function error messages
ERR_CHOMP_MUST_BE_STRING = "chomp: input must be a string, got {type}"
ERR_INDENT_ARGS_MUST_BE_STRINGS = "indent: arguments must be strings"
ERR_JOIN_ARGS_MUST_BE_STRING_AND_LIST = "join: arguments must be string and list/tuple"
ERR_LOWER_MUST_BE_STRING = "lower: input must be a string, got {type}"
ERR_UPPER_MUST_BE_STRING = "upper: input must be a string, got {type}"
ERR_REPLACE_ALL_ARGS_MUST_BE_STRINGS = "replace: all arguments must be strings"
ERR_SPLIT_ARGS_MUST_BE_STRINGS = "split: arguments must be strings"
ERR_STRREV_MUST_BE_STRING = "strrev: input must be a string, got {type}"
ERR_SUBSTR_ARGS_MUST_BE_STRING_NUMBER_NUMBER = "substr: arguments must be string, number, number"
ERR_SUBSTR_OFFSET_MUST_BE_NON_NEGATIVE = "substr: offset must be a non-negative integer"
ERR_SUBSTR_LENGTH_MUST_BE_NON_NEGATIVE_OR_MINUS_ONE = "substr: length must be non-negative or -1"
ERR_TITLE_MUST_BE_STRING = "title: input must be a string, got {type}"
ERR_TRIM_ARGS_MUST_BE_STRINGS = "trim: both arguments must be strings"
ERR_TRIMPREFIX_ARGS_MUST_BE_STRINGS = "trimprefix: both arguments must be strings"
ERR_TRIMSUFFIX_ARGS_MUST_BE_STRINGS = "trimsuffix: both arguments must be strings"
ERR_TRIMSPACE_MUST_BE_STRING = "trimspace: input must be a string, got {type}"

# Regular expression error messages
ERR_REGEX_ARGS_MUST_BE_STRINGS = "regex: both arguments must be strings"
ERR_REGEX_INVALID_EXPRESSION = "regex: invalid regular expression: {error}"
ERR_REGEXALL_ARGS_MUST_BE_STRINGS = "regexall: both arguments must be strings"
ERR_REGEXALL_INVALID_EXPRESSION = "regexall: invalid regular expression: {error}"
ERR_REGEXREPLACE_ALL_ARGS_MUST_BE_STRINGS = "regexreplace: all arguments must be strings"
ERR_REGEXREPLACE_INVALID_EXPRESSION = "regexreplace: invalid regular expression: {error}"

# Collection function error messages
ERR_CHUNKLIST_ARGS_MUST_BE_LIST_AND_NUMBER = "chunklist: arguments must be a list/tuple and a number"
ERR_CHUNKLIST_SIZE_MUST_BE_POSITIVE = "chunklist: size must be a positive number"
ERR_COALESCE_MIN_ONE_ARG = "coalesce must have at least one argument"
ERR_COALESCELIST_NO_NON_EMPTY_LIST = "coalescelist: no non-empty list or tuple found in arguments"
ERR_COMPACT_ARG_MUST_BE_COLLECTION_OF_STRINGS = "compact: argument must be a list, set, or tuple of strings"
ERR_CONCAT_ALL_ARGS_MUST_BE_LISTS = "concat: all arguments must be lists or tuples"
ERR_CONTAINS_COLLECTION_MUST_BE_LIST_SET_TUPLE = (
    "contains: collection must be a list, set, or tuple, got {type}"
)
ERR_DISTINCT_INPUT_MUST_BE_LIST_SET_TUPLE = "distinct: input must be a list, set, or tuple, got {type}"
ERR_DISTINCT_ELEMENT_NOT_HASHABLE = "distinct: element of type {type} is not hashable. Error: {error}"
ERR_ELEMENT_CANNOT_USE_WITH_EMPTY_LIST = "element: cannot use element function with an empty list"
ERR_ELEMENT_COLLECTION_MUST_BE_LIST_TUPLE = "element: collection must be a list or tuple"
ERR_FLATTEN_INPUT_MUST_BE_LIST_SET_TUPLE = "flatten: input must be a list, set, or tuple, got {type}"
ERR_FLATTEN_ALL_ELEMENTS_MUST_BE_LISTS = "flatten: all elements must be lists, sets, or tuples; found {type}"
ERR_HASINDEX_COLLECTION_MUST_BE_LIST_TUPLE_MAP_OBJECT = (
    "hasindex: collection must be a list, tuple, map, or object, got {type}"
)
ERR_INDEX_KEY_DOES_NOT_EXIST = "index: key does not exist in collection"
ERR_KEYS_INPUT_MUST_BE_MAP_OBJECT = "keys: input must be a map or object, got {type}"
ERR_LENGTH_INPUT_MUST_BE_COLLECTION_STRING = "length: input must be a collection or string, got {type}"
ERR_LOOKUP_COLLECTION_MUST_BE_MAP_OBJECT = "lookup: collection must be a map or object"
ERR_MERGE_ALL_ARGS_MUST_BE_MAPS_OBJECTS = "merge: all arguments must be maps or objects"
ERR_REVERSE_INPUT_MUST_BE_LIST_TUPLE = "reverse: input must be a list or tuple"
ERR_SETPRODUCT_ALL_ARGS_MUST_BE_COLLECTIONS = "setproduct: all arguments must be collections"
ERR_SLICE_ARGS_MUST_BE_START_END_NUMBERS = "slice: start and end must be numbers"
ERR_SLICE_INPUT_MUST_BE_LIST_TUPLE = "slice: input must be a list or tuple, got {type}"
ERR_SORT_INPUT_VALUE_NOT_ITERABLE = "sort: input value is not iterable"
ERR_SORT_INPUT_MUST_BE_LIST_SET_TUPLE = "sort: input must be a list, set, or tuple, got {type}"
ERR_SORT_CANNOT_SORT_WITH_NULL_UNKNOWN = (
    "sort: cannot sort list with null or unknown elements at index {index}."
)
ERR_SORT_ELEMENTS_MUST_BE_STRING_NUMBER_BOOL = "sort: elements must be string, number, or bool. Found: {type}"
ERR_VALUES_INPUT_MUST_BE_MAP_OBJECT = "values: input must be a map or object, got {type}"
ERR_ZIPMAP_ARGS_MUST_BE_LISTS = "zipmap: arguments must be lists or tuples"

# Type conversion function error messages
ERR_TOBOOL_CANNOT_CONVERT = "tobool: cannot convert {type} to bool"
ERR_TOSTRING_CANNOT_CONVERT = "tostring: cannot convert {type} to number"

# Codec and JSON error messages
ERR_CSVDECODE_ARG_MUST_BE_STRING = "csvdecode: argument must be a string, got {type}"
ERR_CSVDECODE_FAILED = "csvdecode: failed to decode CSV: {error}"
ERR_JSONDECODE_ARG_MUST_BE_STRING = "jsondecode: argument must be a string, got {type}"
ERR_JSONDECODE_FAILED = "jsondecode: failed to decode JSON: {error}"
ERR_JSONENCODE_FAILED = "jsonencode: failed to encode value: {error}"

# Date and time function error messages
ERR_FORMATDATE_ARGS_MUST_BE_STRINGS = "formatdate: arguments must be strings"
ERR_FORMATDATE_INVALID_TIMESTAMP = "formatdate: invalid timestamp format: {error}"
ERR_TIMEADD_ARGS_MUST_BE_STRINGS = "timeadd: arguments must be strings"
ERR_TIMEADD_INVALID_FORMAT = "timeadd: invalid argument format: {error}"
ERR_INVALID_DURATION_FORMAT = "Invalid duration string format: '{duration_str}'"

# Parsing and validation error messages
ERR_PARSEINT_ARGS_MUST_BE_STRING_NUMBER = "parseint: arguments must be string and number"
ERR_PARSEINT_BASE_INVALID = "parseint: base must be 0 or between 2 and 36, got {base}"
ERR_FAILED_DECODE_DYNAMIC_TYPE_VALIDATION = (
    "Failed to decode dynamic value type spec from JSON during validation"
)
ERR_FAILED_PROCESS_ELEMENT_FOR_SET = "Failed to process element for set: {error}"
ERR_ERROR_AT_STEP = "Error at step {step_num} ({step}): {error}"
ERR_ERROR_AT_TYPE_STEP = "Error at type step {step_num} ({step}): {error}"

# Bytes and capsule function error messages
ERR_BYTESLEN_ARG_MUST_BE_BYTES_CAPSULE = "byteslen: argument must be a Bytes capsule, got {type}"
ERR_BYTESSLICE_ARGS_MUST_BE_BYTES_NUMBER_NUMBER = "bytesslice: arguments must be Bytes capsule, number, number"

# Generic value operation error messages
ERR_GET_OPERATION_NON_MAP_CTYVALUE = "get operation called on non-map CtyValue"
ERR_GET_ATTRIBUTE_REQUIRES_CTYVALUE_OBJECT = "get_attribute requires a CtyValue object"
ERR_ATTRIBUTE_CANNOT_BE_NULL = "Attribute cannot be null"
ERR_ATTRIBUTE_NAME_CANNOT_BE_EMPTY = "Attribute name cannot be empty"
ERR_UNKNOWN_ATTRIBUTES = "Unknown attributes: {attributes}"
ERR_UNKNOWN_OPTIONAL_ATTRIBUTES = "Unknown optional attributes: {attributes}"

# Internal errors
ERR_INTERNAL_CTYVALUE_CTYLIST_NOT_LIST_TUPLE = (
    "Internal error: CtyValue of CtyList type does not wrap a list/tuple, got {type_name}"
)
ERR_INTERNAL_CTYVALUE_CTYMAP_NOT_DICT = (
    "Internal error: CtyValue of CtyMap type does not wrap a dict, got {type_name}"
)
ERR_EXPECTED_CTYVALUE_CTYLIST = "Expected CtyValue with CtyList type, got CtyValue with {type_name}"
ERR_EXPECTED_CTYVALUE_LIST = "Expected CtyValue[CtyList], got {type_name}"
ERR_EXPECTED_LIST_TUPLE_CTYVALUE_LIST = "Expected list, tuple, or CtyValue list, got {type_name}"

# 🌊🪢🔚



#
# SPDX-FileCopyrightText: Copyright (c) provide.io llc. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#


from __future__ import annotations

from typing import Self

from attrs import define
from provide.foundation.config import RuntimeConfig, env_field

from pyvider.cty.config.defaults import ENABLE_TYPE_INFERENCE_CACHE

"""Runtime configuration for pyvider-cty using Foundation config patterns."""


@define
class CtyConfig(RuntimeConfig):
    """Runtime configuration for pyvider-cty.

    Uses Foundation's RuntimeConfig for consistent environment variable loading
    and validation patterns.
    """

    enable_type_inference_cache: bool = env_field(
        env_var="PYVIDER_CTY_ENABLE_TYPE_INFERENCE_CACHE",
        default=ENABLE_TYPE_INFERENCE_CACHE,
    )

    @classmethod
    def get_current(cls) -> Self:
        """Get current configuration from environment variables.

        Returns:
            Current CtyConfig instance loaded from environment
        """
        return cls.from_env(prefix="PYVIDER_CTY")


# 🌊🪢🔚



#
# SPDX-FileCopyrightText: Copyright (c) provide.io llc. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#


from __future__ import annotations

from decimal import Decimal, InvalidOperation
import math
from typing import Any, cast

from pyvider.cty import CtyNumber, CtyString, CtyValue
from pyvider.cty.config.defaults import POSITIVE_BOUNDARY, ZERO_VALUE
from pyvider.cty.exceptions import CtyFunctionError
from pyvider.cty.values.markers import RefinedUnknownValue


def _get_refined_components(
    a: CtyValue[Any], b: CtyValue[Any]
) -> tuple[RefinedUnknownValue, RefinedUnknownValue, Any, Any]:
    """Extract refinement components from two values."""
    ref_a = a.value if isinstance(a.value, RefinedUnknownValue) else RefinedUnknownValue()
    ref_b = b.value if isinstance(b.value, RefinedUnknownValue) else RefinedUnknownValue()
    val_a = a.value if not a.is_unknown else None
    val_b = b.value if not b.is_unknown else None
    return ref_a, ref_b, val_a, val_b


def _propagate_add_refinements(
    ref_a: RefinedUnknownValue, ref_b: RefinedUnknownValue, val_a: Any, val_b: Any
) -> dict[str, Any]:
    """Handle refinement propagation for addition."""
    new_ref: dict[str, Any] = {}
    if val_a is not None:  # a is known, b is refined/unrefined
        if ref_b.number_lower_bound:
            new_ref["number_lower_bound"] = (
                val_a + ref_b.number_lower_bound[0],
                ref_b.number_lower_bound[1],
            )
        if ref_b.number_upper_bound:
            new_ref["number_upper_bound"] = (
                val_a + ref_b.number_upper_bound[0],
                ref_b.number_upper_bound[1],
            )
    elif val_b is not None:  # b is known, a is refined/unrefined
        if ref_a.number_lower_bound:
            new_ref["number_lower_bound"] = (
                ref_a.number_lower_bound[0] + val_b,
                ref_a.number_lower_bound[1],
            )
        if ref_a.number_upper_bound:
            new_ref["number_upper_bound"] = (
                ref_a.number_upper_bound[0] + val_b,
                ref_a.number_upper_bound[1],
            )
    else:  # both are refined/unrefined
        if ref_a.number_lower_bound and ref_b.number_lower_bound:
            new_ref["number_lower_bound"] = (
                ref_a.number_lower_bound[0] + ref_b.number_lower_bound[0],
                ref_a.number_lower_bound[1] and ref_b.number_lower_bound[1],
            )
        if ref_a.number_upper_bound and ref_b.number_upper_bound:
            new_ref["number_upper_bound"] = (
                ref_a.number_upper_bound[0] + ref_b.number_upper_bound[0],
                ref_a.number_upper_bound[1] and ref_b.number_upper_bound[1],
            )
    return new_ref


def _propagate_subtract_refinements(
    ref_a: RefinedUnknownValue, ref_b: RefinedUnknownValue, val_a: Any, val_b: Any
) -> dict[str, Any]:
    """Handle refinement propagation for subtraction."""
    new_ref: dict[str, Any] = {}
    if val_b is not None:
        if ref_a.number_lower_bound:
            new_ref["number_lower_bound"] = (
                ref_a.number_lower_bound[0] - val_b,
                ref_a.number_lower_bound[1],
            )
        if ref_a.number_upper_bound:
            new_ref["number_upper_bound"] = (
                ref_a.number_upper_bound[0] - val_b,
                ref_a.number_upper_bound[1],
            )
    elif val_a is not None:
        if ref_b.number_upper_bound:
            new_ref["number_lower_bound"] = (
                val_a - ref_b.number_upper_bound[0],
                ref_b.number_upper_bound[1],
            )
        if ref_b.number_lower_bound:
            new_ref["number_upper_bound"] = (
                val_a - ref_b.number_lower_bound[0],
                ref_b.number_lower_bound[1],
            )
    else:
        if ref_a.number_lower_bound and ref_b.number_upper_bound:
            new_ref["number_lower_bound"] = (
                ref_a.number_lower_bound[0] - ref_b.number_upper_bound[0],
                ref_a.number_lower_bound[1] and ref_b.number_upper_bound[1],
            )
        if ref_a.number_upper_bound and ref_b.number_lower_bound:
            new_ref["number_upper_bound"] = (
                ref_a.number_upper_bound[0] - ref_b.number_lower_bound[0],
                ref_a.number_upper_bound[1] and ref_b.number_lower_bound[1],
            )
    return new_ref


def _propagate_multiply_refinements(
    ref_a: RefinedUnknownValue, ref_b: RefinedUnknownValue, val_a: Any, val_b: Any
) -> dict[str, Any]:
    """Handle refinement propagation for multiplication."""
    new_ref: dict[str, Any] = {}
    known_val, unknown_ref = (val_a, ref_b) if val_a is not None else (val_b, ref_a)
    if known_val is not None:
        if known_val > POSITIVE_BOUNDARY:
            if unknown_ref.number_lower_bound:
                new_ref["number_lower_bound"] = (
                    unknown_ref.number_lower_bound[0] * known_val,
                    unknown_ref.number_lower_bound[1],
                )
            if unknown_ref.number_upper_bound:
                new_ref["number_upper_bound"] = (
                    unknown_ref.number_upper_bound[0] * known_val,
                    unknown_ref.number_upper_bound[1],
                )
        elif known_val < POSITIVE_BOUNDARY:
            if unknown_ref.number_upper_bound:
                new_ref["number_lower_bound"] = (
                    unknown_ref.number_upper_bound[0] * known_val,
                    unknown_ref.number_upper_bound[1],
                )
            if unknown_ref.number_lower_bound:
                new_ref["number_upper_bound"] = (
                    unknown_ref.number_lower_bound[0] * known_val,
                    unknown_ref.number_lower_bound[1],
                )
    return new_ref


def _propagate_divide_refinements(ref_a: RefinedUnknownValue, val_b: Any) -> dict[str, Any]:
    """Handle refinement propagation for division."""
    new_ref: dict[str, Any] = {}
    if val_b is not None:
        if val_b > POSITIVE_BOUNDARY:
            if ref_a.number_lower_bound:
                new_ref["number_lower_bound"] = (
                    ref_a.number_lower_bound[0] / val_b,
                    ref_a.number_lower_bound[1],
                )
            if ref_a.number_upper_bound:
                new_ref["number_upper_bound"] = (
                    ref_a.number_upper_bound[0] / val_b,
                    ref_a.number_upper_bound[1],
                )
        elif val_b < POSITIVE_BOUNDARY:
            if ref_a.number_upper_bound:
                new_ref["number_lower_bound"] = (
                    ref_a.number_upper_bound[0] / val_b,
                    ref_a.number_upper_bound[1],
                )
            if ref_a.number_lower_bound:
                new_ref["number_upper_bound"] = (
                    ref_a.number_lower_bound[0] / val_b,
                    ref_a.number_lower_bound[1],
                )
    return new_ref


def _propagate_refined_unknowns(op: str, a: CtyValue[Any], b: CtyValue[Any]) -> CtyValue[Any]:
    """Helper to propagate refinements for binary numeric operations."""
    if not (isinstance(a.value, RefinedUnknownValue) or isinstance(b.value, RefinedUnknownValue)):
        return CtyValue.unknown(CtyNumber())

    ref_a, ref_b, val_a, val_b = _get_refined_components(a, b)

    if op == "add":
        new_ref = _propagate_add_refinements(ref_a, ref_b, val_a, val_b)
    elif op == "subtract":
        new_ref = _propagate_subtract_refinements(ref_a, ref_b, val_a, val_b)
    elif op == "multiply":
        new_ref = _propagate_multiply_refinements(ref_a, ref_b, val_a, val_b)
    elif op == "divide":
        new_ref = _propagate_divide_refinements(ref_a, val_b)
    else:
        new_ref = {}

    return CtyValue.unknown(CtyNumber(), value=RefinedUnknownValue(**new_ref))


def add(a: CtyValue[Any], b: CtyValue[Any]) -> CtyValue[Any]:
    if not isinstance(a.type, CtyNumber) or not isinstance(b.type, CtyNumber):
        raise CtyFunctionError("add: arguments must be numbers")
    if a.is_null or b.is_null:
        return CtyValue.unknown(CtyNumber())
    if a.is_unknown or b.is_unknown:
        return _propagate_refined_unknowns("add", a, b)
    a_val = cast(Decimal, a.value)
    b_val = cast(Decimal, b.value)
    return CtyNumber().validate(a_val + b_val)


def subtract(a: CtyValue[Any], b: CtyValue[Any]) -> CtyValue[Any]:
    if not isinstance(a.type, CtyNumber) or not isinstance(b.type, CtyNumber):
        raise CtyFunctionError("subtract: arguments must be numbers")
    if a.is_null or b.is_null:
        return CtyValue.unknown(CtyNumber())
    if a.is_unknown or b.is_unknown:
        return _propagate_refined_unknowns("subtract", a, b)
    a_val = cast(Decimal, a.value)
    b_val = cast(Decimal, b.value)
    return CtyNumber().validate(a_val - b_val)


def multiply(a: CtyValue[Any], b: CtyValue[Any]) -> CtyValue[Any]:
    if not isinstance(a.type, CtyNumber) or not isinstance(b.type, CtyNumber):
        raise CtyFunctionError("multiply: arguments must be numbers")
    if a.is_null or b.is_null:
        return CtyValue.unknown(CtyNumber())
    if (not a.is_unknown and a.value == ZERO_VALUE) or (not b.is_unknown and b.value == ZERO_VALUE):
        return CtyNumber().validate(ZERO_VALUE)
    if a.is_unknown or b.is_unknown:
        return _propagate_refined_unknowns("multiply", a, b)
    a_val = cast(Decimal, a.value)
    b_val = cast(Decimal, b.value)
    return CtyNumber().validate(a_val * b_val)


def divide(a: CtyValue[Any], b: CtyValue[Any]) -> CtyValue[Any]:
    if not isinstance(a.type, CtyNumber) or not isinstance(b.type, CtyNumber):
        raise CtyFunctionError("divide: arguments must be numbers")
    if a.is_null or b.is_null:
        return CtyValue.unknown(CtyNumber())
    if not b.is_unknown and b.value == ZERO_VALUE:
        raise CtyFunctionError("divide by zero")
    if a.is_unknown or b.is_unknown:
        return _propagate_refined_unknowns("divide", a, b)
    a_val = cast(Decimal, a.value)
    b_val = cast(Decimal, b.value)
    return CtyNumber().validate(a_val / b_val)


def modulo(a: CtyValue[Any], b: CtyValue[Any]) -> CtyValue[Any]:
    if not isinstance(a.type, CtyNumber) or not isinstance(b.type, CtyNumber):
        raise CtyFunctionError("modulo: arguments must be numbers")
    if a.is_null or a.is_unknown or b.is_null or b.is_unknown:
        return CtyValue.unknown(CtyNumber())
    if b.value == ZERO_VALUE:
        raise CtyFunctionError("modulo by zero")
    a_val = cast(Decimal, a.value)
    b_val = cast(Decimal, b.value)
    return CtyNumber().validate(Decimal(str(math.fmod(float(a_val), float(b_val)))))


def negate(a: CtyValue[Any]) -> CtyValue[Any]:
    if not isinstance(a.type, CtyNumber):
        raise CtyFunctionError("negate: argument must be a number")
    if a.is_null:
        return CtyValue.null(CtyNumber())
    if a.is_unknown:
        if isinstance(a.value, RefinedUnknownValue):
            ref = a.value
            new_ref = {}
            if ref.number_upper_bound:
                new_ref["number_lower_bound"] = (
                    -ref.number_upper_bound[0],
                    ref.number_upper_bound[1],
                )
            if ref.number_lower_bound:
                new_ref["number_upper_bound"] = (
                    -ref.number_lower_bound[0],
                    ref.number_lower_bound[1],
                )
            return (
                CtyValue.unknown(CtyNumber(), value=RefinedUnknownValue(**new_ref))  # type: ignore[arg-type]
                if new_ref
                else CtyValue.unknown(CtyNumber())
            )
        return CtyValue.unknown(CtyNumber())
    a_val = cast(Decimal, a.value)
    return CtyNumber().validate(-a_val)


def abs_fn(input_val: CtyValue[Any]) -> CtyValue[Any]:
    if not isinstance(input_val.type, CtyNumber):
        raise CtyFunctionError(f"abs: input must be a number, got {input_val.type.ctype}")
    if input_val.is_null:
        return CtyValue.null(CtyNumber())
    if input_val.is_unknown:
        if isinstance(input_val.value, RefinedUnknownValue):
            ref = input_val.value
            new_ref = {}
            lower, upper = ref.number_lower_bound, ref.number_upper_bound
            if lower and upper:
                l_val, l_inc = lower
                u_val, u_inc = upper
                if l_val >= 0:
                    return input_val
                if u_val <= 0:
                    new_ref["number_lower_bound"] = (-u_val, u_inc)
                    new_ref["number_upper_bound"] = (-l_val, l_inc)
                else:
                    new_ref["number_lower_bound"] = (Decimal(0), True)
                    new_upper_val = max(abs(l_val), abs(u_val))
                    new_upper_inc = l_inc if abs(l_val) >= abs(u_val) else u_inc
                    new_ref["number_upper_bound"] = (new_upper_val, new_upper_inc)
            elif lower and lower[0] >= 0:
                return input_val
            elif upper and upper[0] <= 0:
                new_ref["number_lower_bound"] = (-upper[0], upper[1])
            return (
                CtyValue.unknown(CtyNumber(), value=RefinedUnknownValue(**new_ref))  # type: ignore[arg-type]
                if new_ref
                else CtyValue.unknown(CtyNumber())
            )
        return CtyValue.unknown(CtyNumber())
    val = cast(Decimal, input_val.value)
    return CtyNumber().validate(abs(val))


def ceil_fn(input_val: CtyValue[Any]) -> CtyValue[Any]:
    if not isinstance(input_val.type, CtyNumber):
        raise CtyFunctionError(f"ceil: input must be a number, got {input_val.type.ctype}")
    if input_val.is_null or input_val.is_unknown:
        return input_val
    val = cast(Decimal, input_val.value)
    return CtyNumber().validate(Decimal(math.ceil(val)))


def floor_fn(input_val: CtyValue[Any]) -> CtyValue[Any]:
    if not isinstance(input_val.type, CtyNumber):
        raise CtyFunctionError(f"floor: input must be a number, got {input_val.type.ctype}")
    if input_val.is_null or input_val.is_unknown:
        return input_val
    val = cast(Decimal, input_val.value)
    return CtyNumber().validate(Decimal(math.floor(val)))


def log_fn(num_val: CtyValue[Any], base_val: CtyValue[Any]) -> CtyValue[Any]:
    if not isinstance(num_val.type, CtyNumber) or not isinstance(base_val.type, CtyNumber):
        raise CtyFunctionError("log: arguments must be numbers")
    if num_val.is_null or num_val.is_unknown or base_val.is_null or base_val.is_unknown:
        return CtyValue.unknown(CtyNumber())
    num = cast(Decimal, num_val.value)
    base = cast(Decimal, base_val.value)
    if num <= 0:
        raise CtyFunctionError(f"log: number must be positive, got {num}")
    if base <= 0:
        raise CtyFunctionError(f"log: base must be positive, got {base}")
    if base == 1:
        raise CtyFunctionError("log: base cannot be 1")
    try:
        result = Decimal(str(math.log(float(num), float(base))))
        return CtyNumber().validate(result)
    except ValueError as e:
        raise CtyFunctionError(f"log: math domain error: {e}") from e


def pow_fn(num_val: CtyValue[Any], power_val: CtyValue[Any]) -> CtyValue[Any]:
    if not isinstance(num_val.type, CtyNumber) or not isinstance(power_val.type, CtyNumber):
        raise CtyFunctionError("pow: arguments must be numbers")
    if num_val.is_null or num_val.is_unknown or power_val.is_null or power_val.is_unknown:
        return CtyValue.unknown(CtyNumber())
    try:
        num = cast(Decimal, num_val.value)
        power = cast(Decimal, power_val.value)
        result = num**power
        return CtyNumber().validate(result)
    except InvalidOperation as e:
        raise CtyFunctionError(f"pow: invalid operation: {e}") from e


def signum_fn(input_val: CtyValue[Any]) -> CtyValue[Any]:
    if not isinstance(input_val.type, CtyNumber):
        raise CtyFunctionError(f"signum: input must be a number, got {input_val.type.ctype}")
    if input_val.is_null or input_val.is_unknown:
        return input_val
    val = cast(Decimal, input_val.value)
    if val < 0:
        return CtyNumber().validate(Decimal("-1"))
    if val > 0:
        return CtyNumber().validate(Decimal("1"))
    return CtyNumber().validate(Decimal("0"))


def parseint_fn(str_val: CtyValue[Any], base_val: CtyValue[Any]) -> CtyValue[Any]:
    if not isinstance(str_val.type, CtyString) or not isinstance(base_val.type, CtyNumber):
        raise CtyFunctionError("parseint: arguments must be string and number")
    if str_val.is_null or base_val.is_null:
        return CtyValue.null(CtyNumber())
    if str_val.is_unknown or base_val.is_unknown:
        return CtyValue.unknown(CtyNumber())
    s = cast(str, str_val.value)
    base = int(cast(Decimal, base_val.value))
    if not (base == 0 or 2 <= base <= 36):
        raise CtyFunctionError(f"parseint: base must be 0 or between 2 and 36, got {base}")
    try:
        parsed_int = int(s, base)
        return CtyNumber().validate(Decimal(parsed_int))
    except (ValueError, TypeError):
        return CtyValue.null(CtyNumber())


def int_fn(val: CtyValue[Any]) -> CtyValue[Any]:
    if not isinstance(val.type, CtyNumber):
        raise CtyFunctionError(f"int: argument must be a number, got {val.type.ctype}")
    if val.is_null or val.is_unknown:
        return val
    val_decimal = cast(Decimal, val.value)
    return CtyNumber().validate(Decimal(int(val_decimal)))


# 🌊🪢🔚



#
# SPDX-FileCopyrightText: Copyright (c) provide.io llc. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#


from __future__ import annotations

from typing import Any

from pyvider.cty.config.defaults import (
    ERR_CANNOT_INFER_FROM_CTY_TYPE,
    ERR_CANNOT_INFER_FROM_CTY_VALUE,
)

# pyvider-cty/src/pyvider/cty/conversion/_utils.py
"""Internal conversion utilities to avoid circular dependencies."""


def _attrs_to_dict_safe(inst: Any) -> dict[str, Any]:
    """
    Safely converts an attrs instance to a dict, raising TypeError for CTY
    framework types to prevent accidental misuse during type inference.
    """
    # Local imports to prevent circular dependencies at module load time.
    from pyvider.cty.types import CtyType
    from pyvider.cty.values import CtyValue

    if isinstance(inst, CtyType):
        error_message = ERR_CANNOT_INFER_FROM_CTY_TYPE.format(type_name=type(inst).__name__)
        raise TypeError(error_message)
    if isinstance(inst, CtyValue):
        error_message = ERR_CANNOT_INFER_FROM_CTY_VALUE.format(type_name=type(inst).__name__)
        raise TypeError(error_message)

    res = {}
    # Use getattr to safely access __attrs_attrs__ which may not exist.
    for a in getattr(type(inst), "__attrs_attrs__", []):
        res[a.name] = getattr(inst, a.name)
    return res


# 🌊🪢🔚



#
# SPDX-FileCopyrightText: Copyright (c) provide.io llc. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#


from __future__ import annotations

import csv
import io
import json
from typing import Any, cast

from pyvider.cty import CtyDynamic, CtyList, CtyObject, CtyString, CtyValue
from pyvider.cty.conversion import cty_to_native
from pyvider.cty.exceptions import CtyFunctionError


def jsonencode(val: CtyValue[Any]) -> CtyValue[Any]:
    if val.is_unknown:
        return CtyValue.unknown(CtyString())
    try:
        native_val = cty_to_native(val)
        return CtyString().validate(json.dumps(native_val))
    except Exception as e:
        raise CtyFunctionError(f"jsonencode: failed to encode value: {e}") from e


def jsondecode(val: CtyValue[Any]) -> CtyValue[Any]:
    if not isinstance(val.type, CtyString):
        raise CtyFunctionError(f"jsondecode: argument must be a string, got {val.type.ctype}")
    if val.is_unknown or val.is_null:
        return CtyValue.unknown(CtyDynamic())
    try:
        val_str = cast(str, val.value)
        native_val = json.loads(val_str)
        result: CtyValue[Any] = CtyDynamic().validate(native_val)
        return result
    except json.JSONDecodeError as e:
        raise CtyFunctionError(f"jsondecode: failed to decode JSON: {e}") from e


def csvdecode(val: CtyValue[Any]) -> CtyValue[Any]:
    if not isinstance(val.type, CtyString):
        raise CtyFunctionError(f"csvdecode: argument must be a string, got {val.type.ctype}")
    if val.is_unknown or val.is_null:
        return CtyValue.unknown(CtyList(element_type=CtyObject({})))

    val_str = cast(str, val.value)
    f = io.StringIO(val_str)
    try:
        # The csv module can raise csv.Error for malformed data
        reader = csv.DictReader(f)
        rows = list(reader)
        result: CtyValue[Any] = CtyList(element_type=CtyDynamic()).validate(rows)
        return result
    except Exception as e:
        raise CtyFunctionError(f"csvdecode: failed to decode CSV: {e}") from e


# 🌊🪢🔚



#
# SPDX-FileCopyrightText: Copyright (c) provide.io llc. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#


from __future__ import annotations

from typing import TYPE_CHECKING, Any, ClassVar, Generic, TypeVar, cast, final

from attrs import define, field

from pyvider.cty.exceptions import CtyListValidationError, CtyValidationError
from pyvider.cty.path import CtyPath, IndexStep
from pyvider.cty.types.base import CtyType
from pyvider.cty.types.structural import CtyDynamic
from pyvider.cty.validation.recursion import with_recursion_detection
from pyvider.cty.values import CtyValue

if TYPE_CHECKING:
    pass

T = TypeVar("T")


@final
@define(frozen=True, slots=True)
class CtyList(CtyType[tuple[T, ...]], Generic[T]):
    ctype: ClassVar[str] = "list"
    _type_order: ClassVar[int] = 5
    element_type: CtyType[T] = field(kw_only=True)

    def __attrs_post_init__(self) -> None:
        if not isinstance(self.element_type, CtyType):
            raise CtyListValidationError(
                f"Expected CtyType for element_type, got {type(self.element_type).__name__}"
            )

    @with_recursion_detection
    def validate(self, value: object) -> CtyValue[tuple[T, ...]]:  # noqa: C901
        from pyvider.cty.values import CtyValue

        if isinstance(value, CtyValue):
            if self.equal(value.type) and isinstance(value.value, tuple):
                return cast(CtyValue[tuple[T, ...]], value)  # Fast path for already-validated values
            if value.is_null:
                return CtyValue.null(self)
            if value.is_unknown:
                return CtyValue.unknown(self)
            value = value.value

        if value is None:
            return CtyValue.null(self)

        # Check for UnrefinedUnknownValue which can slip through if a CtyValue wrapper is removed
        from pyvider.cty.values.markers import UnrefinedUnknownValue

        if isinstance(value, UnrefinedUnknownValue):
            raise CtyListValidationError(
                "Cannot use unknown/computed value for list parameter. "
                "This value won't be known until apply time, but is needed during validation. "
                "Possible causes:\n"
                "  - Circular reference (e.g., data source referencing itself)\n"
                "  - Using output from another resource/data source that hasn't been created yet\n"
                "  - Dynamic values in function calls during validation\n"
                "Hint: Check for self-references or values that depend on resources not yet created."
            )

        if isinstance(value, list | tuple | set | frozenset):
            value_collection = cast(list[object] | tuple[object, ...] | set[object] | frozenset[object], value)
            raw_list_to_validate: list[object] = list(value_collection)
        else:
            raise CtyListValidationError(f"Expected list, tuple, or CtyValue list, got {type(value).__name__}")

        if not isinstance(raw_list_to_validate, list | tuple):
            raise CtyListValidationError(
                f"Value to validate is not a list or tuple, but {type(raw_list_to_validate).__name__}"
            )

        validated_elements: list[CtyValue[T]] = []
        for i, item in enumerate(raw_list_to_validate):
            if item is None and not isinstance(self.element_type, CtyDynamic):
                raise CtyListValidationError(
                    f"List elements cannot be null for element type {self.element_type.ctype}",
                    path=CtyPath(steps=[IndexStep(i)]),
                )
            try:
                validated_item = self.element_type.validate(item)
                validated_elements.append(validated_item)
            except CtyValidationError as e:
                new_path = CtyPath(steps=[IndexStep(i)] + (e.path.steps if e.path else []))
                raise CtyListValidationError(e.message, value=item, path=new_path, original_exception=e) from e

        is_unknown = any(v.is_unknown for v in validated_elements)
        return CtyValue(vtype=self, value=tuple(validated_elements), is_unknown=is_unknown)

    def element_at(self, container: object, index: int) -> CtyValue[T]:
        from pyvider.cty.values import CtyValue

        if isinstance(container, CtyValue):
            if not isinstance(container.type, CtyList):
                raise CtyListValidationError(
                    f"Expected CtyValue with CtyList type, got CtyValue with {type(container.type).__name__}"
                )
            if container.is_null:
                raise IndexError(f"Cannot access element at index {index} in a null list.")
            if container.is_unknown:
                return CtyValue.unknown(self.element_type)
            if not isinstance(container.value, list | tuple):
                raise CtyListValidationError(
                    f"Internal error: CtyValue of CtyList type does not wrap a list/tuple, got {type(container.value).__name__}"
                )
            try:
                container_value_seq = cast(list[Any] | tuple[Any, ...], container.value)  # type: ignore[redundant-cast]
                return self.element_type.validate(container_value_seq[index])
            except TypeError as e:
                raise TypeError(f"list indices must be integers or slices, not {type(index).__name__}") from e

        raise CtyListValidationError(f"Expected CtyValue[CtyList], got {type(container).__name__}")

    def equal(self, other: CtyType[Any]) -> bool:
        if not isinstance(other, CtyList):
            return False
        return self.element_type.equal(other.element_type)

    def usable_as(self, other: CtyType[Any]) -> bool:
        from pyvider.cty.types.structural import CtyDynamic

        if isinstance(other, CtyDynamic):
            return True
        if not isinstance(other, CtyList):
            return False
        return self.element_type.usable_as(other.element_type)

    def _to_wire_json(self) -> Any:
        return [self.ctype, self.element_type._to_wire_json()]

    def __str__(self) -> str:
        return f"list({self.element_type})"

    def __repr__(self) -> str:
        return f"CtyList(element_type={self.element_type!r})"


# 🌊🪢🔚



#
# SPDX-FileCopyrightText: Copyright (c) provide.io llc. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#


from __future__ import annotations

from pyvider.cty.conversion.adapter import cty_to_native
from pyvider.cty.conversion.explicit import convert, unify

# pyvider-cty/src/pyvider/cty/conversion/__init__.py
from pyvider.cty.conversion.inference_cache import inference_cache_context, with_inference_cache
from pyvider.cty.conversion.raw_to_cty import infer_cty_type_from_raw
from pyvider.cty.conversion.type_encoder import encode_cty_type_to_wire_json

__all__ = [
    "convert",
    "cty_to_native",
    "encode_cty_type_to_wire_json",
    "infer_cty_type_from_raw",
    "inference_cache_context",
    "unify",
    "with_inference_cache",
]

# 🌊🪢🔚



#
# SPDX-FileCopyrightText: Copyright (c) provide.io llc. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#


from __future__ import annotations

from typing import Any

from pyvider.cty.exceptions.base import CtyError

# pyvider/cty/exceptions/conversion.py
"""
Defines exceptions related to CTY type and value conversions.
"""


class CtyConversionError(CtyError):
    """Base for CTY value or type conversion errors."""

    def __init__(
        self,
        message: str,
        *,
        source_value: object | None = None,
        target_type: object | None = None,
        **kwargs: Any,
    ) -> None:
        """
        Initializes the CtyConversionError.

        Args:
            message: The base error message.
            source_value: The value that was being converted.
            target_type: The intended target type of the conversion.
            **kwargs: Additional foundation error context.
        """
        self.source_value = source_value
        self.target_type = target_type

        # Add rich conversion context
        context = kwargs.setdefault("context", {})
        context["cty.operation"] = "conversion"
        context["cty.error_category"] = "type_conversion"

        # Build message with old format for compatibility
        context_parts = []
        if source_value is not None:
            context_parts.append(f"source_type={type(source_value).__name__}")
            # Also add to foundation context
            context["conversion.source_type"] = type(source_value).__name__
            context["conversion.source_value_type"] = type(source_value).__name__

            # Add value analysis for better debugging
            if hasattr(source_value, "type") and hasattr(source_value, "is_null"):
                context["conversion.source_cty_type"] = str(source_value.type)
                context["conversion.source_is_null"] = source_value.is_null
                if hasattr(source_value, "is_unknown"):
                    context["conversion.source_is_unknown"] = source_value.is_unknown

        if target_type is not None:
            target_name = target_type.__name__ if hasattr(target_type, "__name__") else str(target_type)
            context_parts.append(f"target_type={target_name}")
            context["conversion.target_type"] = target_name
            context["conversion.target_type_str"] = str(target_type)

        if context_parts:
            message = f"{message} ({', '.join(context_parts)})"

        super().__init__(message, **kwargs)

    def _default_code(self) -> str:
        return "CTY_CONVERSION_ERROR"


class CtyTypeConversionError(CtyConversionError):
    """CTY type representation conversion failure."""

    def __init__(
        self,
        message: str,
        *,
        type_name: str | None = None,
        source_value: object | None = None,
        target_type: object | None = None,
        **kwargs: Any,
    ) -> None:
        """
        Initializes the CtyTypeConversionError.

        Args:
            message: The base error message.
            type_name: The name of the CTY type involved in the conversion failure.
            source_value: The value that was being converted.
            target_type: The intended target type of the conversion.
        """
        self.type_name = type_name

        # Add type-specific context
        context = kwargs.setdefault("context", {})
        context["cty.conversion_category"] = "type_representation"

        if type_name:
            context["cty.failing_type"] = type_name
            message = f'CTY Type "{type_name}" representation conversion failed: {message}'

        super().__init__(message, source_value=source_value, target_type=target_type, **kwargs)


class CtyTypeParseError(CtyConversionError):
    """Raised when a CTY type string cannot be parsed."""

    def __init__(self, message: str, type_string: str, **kwargs: Any) -> None:
        self.type_string = type_string

        # Add parsing context
        context = kwargs.setdefault("context", {})
        context["cty.conversion_category"] = "type_parsing"
        context["cty.parse_input"] = str(type_string)[:100]  # Truncate for safety
        context["cty.parse_input_type"] = type(type_string).__name__

        full_message = f"{message}: '{type_string}'"
        super().__init__(full_message, source_value=type_string, **kwargs)


__all__ = ["CtyConversionError", "CtyTypeConversionError", "CtyTypeParseError"]

# 🌊🪢🔚



#
# SPDX-FileCopyrightText: Copyright (c) provide.io llc. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#


from __future__ import annotations

from typing import Any, ClassVar, Generic, TypeVar, cast
import unicodedata

from attrs import define, field

from pyvider.cty.exceptions import (
    CtyMapValidationError,
    CtyTypeMismatchError,
    CtyValidationError,
    InvalidTypeError,
)
from pyvider.cty.path import CtyPath, KeyStep
from pyvider.cty.types.base import CtyType
from pyvider.cty.validation.recursion import with_recursion_detection
from pyvider.cty.values import CtyValue

V = TypeVar("V")


@define(frozen=True, slots=True)
class CtyMap(CtyType[dict[str, V]], Generic[V]):
    ctype: ClassVar[str] = "map"
    _type_order: ClassVar[int] = 6
    element_type: CtyType[V] = field(kw_only=True)

    def __attrs_post_init__(self) -> None:
        if not isinstance(self.element_type, CtyType):
            raise InvalidTypeError(
                f"element_type must be a CtyType instance, got {type(self.element_type).__name__}"
            )

    @with_recursion_detection
    def validate(self, value: object) -> CtyValue[dict[str, V]]:
        if isinstance(value, CtyValue):
            if self.equal(value.type) and isinstance(value.value, dict):
                return cast(CtyValue[dict[str, V]], value)  # Fast path
            if value.is_null:
                return CtyValue.null(self)
            if value.is_unknown:
                return CtyValue.unknown(self)
            value = value.value

        if value is None:
            return CtyValue.null(self)

        if not isinstance(value, dict):
            raise CtyMapValidationError(f"Input must be a dictionary, got {type(value).__name__}.")
        validated_map: dict[str, CtyValue[V]] = {}
        for k, v in value.items():
            if not isinstance(k, str):
                raise CtyMapValidationError(
                    f"Map keys must be strings, but got key of type {type(k).__name__}"
                )

            normalized_key = unicodedata.normalize("NFC", k)

            try:
                validated_map[normalized_key] = self.element_type.validate(v)
            except CtyValidationError as e:
                new_path = CtyPath(steps=[KeyStep(normalized_key)] + (e.path.steps if e.path else []))
                raise CtyMapValidationError(e.message, value=v, path=new_path, original_exception=e) from e

        is_unknown = any(v.is_unknown for v in validated_map.values())
        return CtyValue(vtype=self, value=validated_map, is_unknown=is_unknown)

    def get(
        self,
        map_value: CtyValue[dict[str, V]],
        key: object,
        default: CtyValue[V] | None = None,
    ) -> CtyValue[V]:
        if not isinstance(map_value, CtyValue) or not isinstance(map_value.type, CtyMap):
            raise CtyTypeMismatchError("get operation called on non-map CtyValue")
        if map_value.is_null or map_value.is_unknown:
            return default if default is not None else CtyValue.null(self.element_type)
        internal_dict = map_value.value
        if not isinstance(internal_dict, dict):
            raise CtyMapValidationError(
                f"Internal error: CtyValue of CtyMap type does not wrap a dict, got {type(internal_dict).__name__}"
            )

        normalized_key = unicodedata.normalize("NFC", str(key))
        internal_dict_cast = cast(dict[str, CtyValue[V]], internal_dict)
        result = internal_dict_cast.get(normalized_key)

        if result is not None:
            return self.element_type.validate(result)
        return default if default is not None else CtyValue.null(self.element_type)

    def equal(self, other: CtyType[Any]) -> bool:
        if not isinstance(other, CtyMap):
            return False
        return self.element_type.equal(other.element_type)

    def usable_as(self, other: CtyType[Any]) -> bool:
        from pyvider.cty.types.structural import CtyDynamic

        if isinstance(other, CtyDynamic):
            return True
        if not isinstance(other, CtyMap):
            return False
        return self.element_type.usable_as(other.element_type)

    def _to_wire_json(self) -> Any:
        return [self.ctype, self.element_type._to_wire_json()]

    def __str__(self) -> str:
        return f"map({self.element_type})"


# 🌊🪢🔚
